# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Mako utility logic
"""

import os

from mako.lookup import TemplateLookup

from rattail.files import resource_path


class ResourceTemplateLookup(TemplateLookup):
    """
    This logic was largely copied/inspired from pyramid_mako,
    https://github.com/Pylons/pyramid_mako/blob/main/src/pyramid_mako/__init__.py
    """

    def adjust_uri(self, uri, relativeto):

        # do not adjust "resource path spec" uri
        isabs = os.path.isabs(uri)
        if (not isabs) and (':' in uri):
            return uri

        return super(ResourceTemplateLookup, self).adjust_uri(uri, relativeto)

    def get_template(self, uri):

        # check if uri looks like a "resource path spec"
        isabs = os.path.isabs(uri)
        if (not isabs) and (':' in uri):

            # it does..first use normal logic to try fetching from cache
            try:
                if self.filesystem_checks:
                    return self._check(uri, self._collection[uri])
                else:
                    return self._collection[uri]
            except KeyError as e:

                # but if not already in cache, must convert resource
                # path spec to absolute path on disk, and load that
                path = resource_path(uri)
                if os.path.exists(path):
                    return self._load(path, uri)

        # fallback to normal logic
        return super(ResourceTemplateLookup, self).get_template(uri)



# -*- coding: utf-8; -*-
"""add batch_pos_row.cur_price

Revision ID: e112f75359b3
Revises: 7f025fc530cc
Create Date: 2023-10-18 18:42:24.745734

"""

# revision identifiers, used by Alembic.
revision = 'e112f75359b3'
down_revision = '7f025fc530cc'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_pos_row
    op.add_column('batch_pos_row', sa.Column('cur_price', sa.Numeric(precision=8, scale=3), nullable=True))
    op.add_column('batch_pos_row', sa.Column('cur_price_type', sa.Integer(), nullable=True))
    op.add_column('batch_pos_row', sa.Column('cur_price_start', sa.DateTime(), nullable=True))
    op.add_column('batch_pos_row', sa.Column('cur_price_end', sa.DateTime(), nullable=True))
    op.add_column('batch_pos_row', sa.Column('txn_price_adjusted', sa.Boolean(), nullable=True))


def downgrade():

    # batch_pos_row
    op.drop_column('batch_pos_row', 'txn_price_adjusted')
    op.drop_column('batch_pos_row', 'cur_price_end')
    op.drop_column('batch_pos_row', 'cur_price_start')
    op.drop_column('batch_pos_row', 'cur_price_type')
    op.drop_column('batch_pos_row', 'cur_price')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data models for sales
"""

import sqlalchemy as sa

from .core import Base, uuid_column


class Tender(Base):
    """
    Represents a tender for taking payment, or tracking thereof.
    """
    __tablename__ = 'tender'
    __versioned__ = {}

    uuid = uuid_column()

    code = sa.Column(sa.String(length=10), nullable=True, doc="""
    Unique code for the tender.
    """)

    name = sa.Column(sa.String(length=100), nullable=True, doc="""
    Common name for the tender.
    """)

    notes = sa.Column(sa.Text(), nullable=True, doc="""
    Extra notes to describe the tender.
    """)

    is_cash = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating this tender represents "cash" or equivalent.  This
    is mostly used in conjunction with :attr:`allow_cash_back` to
    determine "cash back" behavior at POS.
    """)

    is_foodstamp = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating this tender represents "food stamps" or
    equivalent.  This is mostly used to calculate FS-eligible balance
    at POS.
    """)

    allow_cash_back = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the customer is allowed to "over-pay" for the
    transaction using this tender, with the intention of getting cash
    back.

    Tenders which do *not* have this flag set, are restricted such
    that the tender amount must be less than or equal to the current
    transaction balance.
    """)

    kick_drawer = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the drawer should be kicked open when a
    transaction is finalized which includes this tender.
    """)

    disabled = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the tender is disabled, and should not be used.
    """)

    def __str__(self):
        return str(self.name or '')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Luigi task commands
"""

from __future__ import unicode_literals, absolute_import

import sys

from rattail.commands import Subcommand, date_argument


class Overnight(Subcommand):
    """
    Launch an overnight task for Luigi
    """
    name = 'overnight'
    description = __doc__.strip()

    def add_parser_args(self, parser):

        parser.add_argument('--task_key', '-k',
                            help="Config key for the overnight task to be launched.")

        parser.add_argument('--date', type=date_argument,
                            help="Date for which overnight task should be "
                            "launched.  Defaults to yesterday.")

        parser.add_argument('--email-if-empty', action='store_true', default=True,
                            help="Send email even if task produces no output.")
        parser.add_argument('--no-email-if-empty', action='store_false', dest='email_if_empty',
                            help="Send email only if task produces output.")

        parser.add_argument('--email-key',
                            help="Config key for email settings, to be used in "
                            "determining recipients etc.")

        parser.add_argument('--wait', action='store_true', default=True,
                            help="Run the task in-process, and wait for it to "
                            "finish.  This is the default behavior.")
        parser.add_argument('--no-wait', action='store_false', dest='wait',
                            help="Do not wait for the task to finish.  This will "
                            "cause the command to run via the `at` utility, i.e. "
                            "scheduled to start within the next minute.")

        parser.add_argument('--purge-settings', action='store_true',
                            help="Instead of running a task, remove all settings "
                            "from the DB, for all overnight tasks.")

        parser.add_argument('--dry-run', action='store_true',
                            help="Log the final command for the task, but do not "
                            "actually run it.")

    def run(self, args):
        self.luigi_handler = self.app.get_luigi_handler()
        if args.purge_settings:
            self.purge_settings(args)
        else:
            self.launch_task(args)

    def purge_settings(self, args):
        session = self.make_session()
        self.luigi_handler.purge_overnight_settings(session)
        self.finalize_session(session, dry_run=args.dry_run)

    def launch_task(self, args):
        key = args.task_key
        task = self.luigi_handler.get_overnight_task(key)
        if not task:
            self.stderr.write("overnight task not found for key: {}\n".format(key))
            sys.exit(1)

        date = args.date or self.app.yesterday()
        self.luigi_handler.launch_overnight_task(task, date,
                                                 email_if_empty=args.email_if_empty,
                                                 email_key=args.email_key,
                                                 wait=args.wait,
                                                 dry_run=args.dry_run)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Upgrade handlers
"""

import os
import shutil
import subprocess
import logging
import warnings

from rattail.app import GenericHandler


log = logging.getLogger(__name__)


class UpgradeHandler(GenericHandler):
    """
    Base class and default implementation for upgrade handlers.
    """

    def get_all_systems(self, **kwargs):
        """
        Returns a list of all "systems" available for upgrade.
        """
        systems = []
        keys = self.config.getlist('rattail.upgrades', 'systems',
                                   default=[])
        for key in keys:
            label = self.config.get('rattail.upgrades',
                                    'system.{}.label'.format(key))
            command = self.config.get('rattail.upgrades',
                                      'system.{}.command'.format(key))
            systems.append({'key': key,
                            'label': label,
                            'command': command})

        systems.sort(key=lambda s: s['label'])

        systems.insert(0, {
            'key': 'rattail',
            'label': self.config.app_title(),
            'command': self.config.get('rattail.upgrades', 'command'),
        })

        return systems

    def get_system(self, key, require=False, **kwargs):
        """
        Returns the "system" record for the given key.
        """
        systems = self.get_all_systems(**kwargs)
        for system in systems:
            if system['key'] == key:
                return system
        if require:
            raise KeyError("No system info found for key: {}".format(key))

    def executable(self, upgrade):
        """
        This method should return a boolean indicating whether or not execution
        should be allowed for the upgrade, given its current condition.  The
        default simply returns ``True`` unless the upgrade has already been
        executed.
        """
        if upgrade is None:
            return True
        return not bool(upgrade.executed)

    def mark_executing(self, upgrade):
        upgrade.executing = True
        upgrade.status_code = self.enum.UPGRADE_STATUS_EXECUTING

    def do_execute(self, upgrade, user, **kwargs):
        """
        Perform all steps needed to fully execute the given upgrade.

        Callers should use this method; you can override
        :meth:`execute()` to customize execution logic.
        """
        # execute proper
        success = self.execute(upgrade, user, **kwargs)

        # declare the upgrade no longer executing
        upgrade.executing = False
        upgrade.executed = self.app.make_utc()
        upgrade.executed_by = user

        # set upgrade status, email key
        if success:
            upgrade.status_code = self.enum.UPGRADE_STATUS_SUCCEEDED
            email_key = 'upgrade_success'
        else:
            upgrade.status_code = self.enum.UPGRADE_STATUS_FAILED
            email_key = 'upgrade_failure'

        # figure out url to the upgrade, if we can
        url = self.config.get('tailbone', 'url.upgrade')
        if not url:
            url = self.config.base_url()
            if url:
                url = '{}/upgrades/{{uuid}}'.format(url)
        if not url:
            url = '#'

        # send appropriate email
        system = self.get_system(upgrade.system or 'rattail')
        self.app.send_email(email_key, {
            'upgrade': upgrade,
            'system_title': system['label'],
            'upgrade_url': url.format(uuid=upgrade.uuid),
        })

    def execute(self, upgrade, user, progress=None, **kwargs):
        """
        Execute the given upgrade, as the given user.
        """
        # record pre-upgrade status
        before_path = self.config.upgrade_filepath(upgrade.uuid,
                                                   filename='requirements.before.txt',
                                                   makedirs=True)
        self.record_requirements_snapshot(before_path)

        # get stdout/stderr file paths
        stdout_path = self.config.upgrade_filepath(upgrade.uuid,
                                                   filename='stdout.log')
        stderr_path = self.config.upgrade_filepath(upgrade.uuid,
                                                   filename='stderr.log')

        # figure out the upgrade command
        if upgrade.system:
            system = self.get_system(upgrade.system, require=True)
            cmd = system['command']
            if not cmd:
                raise ValueError("No command defined for system: {}".format(upgrade.system))
        else:
            cmd = self.get_system('rattail')['command']
        cmd = self.config.parse_list(cmd)

        # run the upgrade command
        log.debug("will run upgrade command: %s", cmd)
        with open(stdout_path, 'wb') as stdout:
            with open(stderr_path, 'wb') as stderr:
                upgrade.exit_code = subprocess.call(cmd, stdout=stdout, stderr=stderr)
        logger = log.warning if upgrade.exit_code != 0 else log.debug
        logger("upgrade command exit code was: %s", upgrade.exit_code)

        # record post-upgrade status
        after_path = self.config.upgrade_filepath(upgrade.uuid, filename='requirements.after.txt')
        self.record_requirements_snapshot(after_path)

        # success as boolean
        return upgrade.exit_code == 0

    def record_requirements_snapshot(self, path):
        pip = self.get_pip_path()
        logpath = os.path.join(self.config.workdir(), 'pip.log')

        kwargs = {}
        suppress_stderr = self.config.getbool('rattail.upgrades', 'suppress_pip_freeze_stderr',
                                              default=False, usedb=False)
        if suppress_stderr:
            stderr = open('/dev/null', 'w')
            kwargs['stderr'] = stderr

        with open(path, 'wb') as stdout:
            subprocess.call([pip, '--log', logpath, 'freeze'], stdout=stdout, **kwargs)

        if suppress_stderr:
            stderr.close()

    def get_pip_path(self):
        path = os.path.join(self.config.appdir(), os.pardir, 'bin', 'pip')
        return os.path.abspath(path)

    def delete_files(self, upgrade):
        """
        Delete all data files for the given upgrade.
        """
        path = self.config.upgrade_filepath(upgrade.uuid)
        if os.path.exists(path):
            shutil.rmtree(path)


def get_upgrade_handler(config, default=None):
    """
    Returns an upgrade handler object.
    """
    warnings.warn("get_upgrade_handler() function is deprecated; please "
                  "use AppHandler.get_upgrade_handler() instead",
                  DeprecationWarning, stacklevel=2)

    app = config.get_app()
    return app.get_upgrade_handler(default=default)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Product-related commands
"""

import logging

from rattail.commands.core import Subcommand, date_argument


log = logging.getLogger(__name__)


class UpdateCosts(Subcommand):
    """
    Update (move future to current) costs for all products
    """
    name = 'update-costs'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def run(self, args):
        from rattail.db.continuum import versioning_manager

        session = self.make_session()
        model = self.model
        user = self.get_runas_user(session)
        session.set_continuum_user(user)
        products_handler = self.app.get_products_handler()

        # TODO: even if this works, it seems heavy-handed...
        # (also it *doesn't* work if ran before setting continuum user)
        uow = versioning_manager.unit_of_work(session)
        transaction = uow.create_transaction(session)
        transaction.meta = {'comment': "make future costs become current"}

        now = self.app.make_utc()
        future_costs = session.query(model.ProductFutureCost)\
                              .filter(model.ProductFutureCost.starts <= now)\
                              .all()
        log.info("found %s future costs which should become current", len(future_costs))

        def move(future, i):
            products_handler.make_future_cost_current(future)

        self.progress_loop(move, future_costs,
                           message="Making future costs become current")

        self.finalize_session(session, dry_run=args.dry_run)



# -*- coding: utf-8 -*-
"""add product.organic

Revision ID: 3f5be78f0f18
Revises: 3ef0fbcfb42e
Create Date: 2015-02-25 00:12:42.696926

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '3f5be78f0f18'
down_revision = u'3ef0fbcfb42e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product.organic
    op.add_column('product', sa.Column('organic', sa.Boolean(), nullable=True))
    product = sa.sql.table('product', sa.sql.column('organic'))
    op.execute(product.update().values({'organic': False}))
    op.alter_column('product', 'organic', nullable=False,
                    existing_type=sa.Boolean())
    op.add_column('product_version', sa.Column('organic', sa.Boolean(), nullable=True))


def downgrade():

    # product.organic
    op.drop_column('product_version', 'organic')
    op.drop_column('product', 'organic')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Telemetry logic
"""

import os
import re
import subprocess
import warnings

from rattail.app import GenericHandler
from rattail.config import ConfigProfile
from rattail.tailbone import TailboneAPIClient


class TelemetryHandler(GenericHandler):
    """
    Handler for telemetry data
    """

    def get_profile(self, profile):
        if isinstance(profile, TelemetryProfile):
            return profile

        return TelemetryProfile(self.config, profile or 'default')

    def collect_all_data(self, profile=None, **kwargs):
        data = {}
        profile = self.get_profile(profile)

        for key in profile.collect_keys:
            collector = getattr(self, 'collect_data_{}'.format(key))
            data[key] = collector(profile=profile)

        self.normalize_errors(data)
        return data

    def collect_data_composer(self, profile, **kwargs):
        data = {}
        errors = []

        # executable
        composer = profile._config_string('collect.composer.executable')
        if not composer:
            composer = self.app.get_composer_executable()
        data['executable'] = composer

        # release
        try:
            output = subprocess.check_output([composer, '--version'])
        except subprocess.CalledProcessError:
            errors.append("Failed to execute `composer --version`")
        else:
            output = output.decode('utf_8').strip()
            data['release_full'] = output
            match = re.match(r'^Composer version (\d+\.\d+\.\d+)\s', output)
            if match:
                data['release_version'] = match.group(1)
            else:
                errors.append("Failed to parse Composer version")

        if errors:
            data['errors'] = errors
        return data

    # TODO: this probably belongs in rattail-corepos but we can move
    # it if/when telemetry is refactored for more dynamic profiles
    def collect_data_corepos(self, profile, **kwargs):
        data = {}
        errors = []

        # path (required for others to work)
        path = profile._config_string('collect.corepos.path')
        if not path:
            errors.append("No path configured!")
        else:
            data['path'] = path

            # owner
            try:
                output = subprocess.check_output([
                    'bash', '-c',
                    "ls -ld '{}' | cut -d ' ' -f 3,4".format(path)])
            except subprocess.CalledProcessError:
                errors.append("Failed to check path ownership")
            else:
                output = output.decode('utf_8').strip()
                data['owner'] = output

            # git_origin
            try:
                output = subprocess.check_output([
                    'bash', '-c',
                    "cd {} && git remote get-url origin".format(path)])
            except subprocess.CalledProcessError:
                errors.append("Failed to execute `git remote`")
            else:
                output = output.decode('utf_8').strip()
                data['git_origin'] = output

            # git_described
            try:
                output = subprocess.check_output([
                    'bash', '-c',
                    "cd {} && git describe --tags".format(path)])
            except subprocess.CalledProcessError:
                errors.append("Failed to execute `git describe`")
            else:
                output = output.decode('utf_8').strip()
                data['git_described'] = output

            # git_status
            try:
                output = subprocess.check_output([
                    'bash', '-c',
                    "cd {} && git status --branch --porcelain".format(path)])
            except subprocess.CalledProcessError:
                errors.append("Failed to execute `git status`")
            else:
                output = output.decode('utf_8').strip()
                data['git_status'] = output

            # composer = self.app.get_composer_executable()

            # # composer_status
            # try:
            #     output = subprocess.check_output([
            #         'bash', '-c',
            #         "cd {} && {} status".format(path, composer)
            #     ], stderr=subprocess.STDOUT)
            # except subprocess.CalledProcessError:
            #     errors.append("Failed to execute `composer status`")
            # else:
            #     output = output.decode('utf_8').strip()
            #     data['composer_status'] = output

            # # composer_validate
            # try:
            #     output = subprocess.check_output([
            #         'bash', '-c',
            #         "cd {} && {} validate".format(path, composer)
            #     ], stderr=subprocess.STDOUT)
            # except subprocess.CalledProcessError:
            #     errors.append("Failed to execute `composer validate`")
            # else:
            #     output = output.decode('utf_8').strip()
            #     data['composer_validated'] = output

        if errors:
            data['errors'] = errors
        return data

    def collect_data_git(self, profile, **kwargs):
        data = {}
        errors = []

        # executable
        git = profile._config_string('collect.git.executable',
                                     default='git')
        data['executable'] = git

        # release
        try:
            output = subprocess.check_output([git, '--version'])
        except subprocess.CalledProcessError:
            errors.append("Failed to execute `git --version`")
        else:
            output = output.decode('utf_8').strip()
            data['release_full'] = output
            match = re.match(r'^git version (\d+\.\d+\.\d+)', output)
            if match:
                data['release_version'] = match.group(1)
            else:
                errors.append("Failed to parse Git version")

        if errors:
            data['errors'] = errors
        return data

    def collect_data_mysql(self, profile, **kwargs):
        data = {}
        errors = []

        # release
        try:
            output = subprocess.check_output(['mysql', '--version'])
        except subprocess.CalledProcessError:
            errors.append("Failed to execute `mysql --version`")
        else:
            output = output.decode('utf_8').strip()
            data['release_full'] = output

            if 'mariadb' in output.lower():
                data['release_id'] = 'mariadb'
                match = re.match(r'^.*\s(\d+(?:\.\d+)+)-MariaDB', output)
                if match:
                    data['release_version'] = match.group(1)
                else:
                    errors.append("Failed to parse MariaDB version")
            else:
                data['release_id'] = 'mysql'
                match = re.match(r'^mysql\s+Ver (\d+(?:\.\d+)+)-', output)
                if match:
                    data['release_version'] = match.group(1)
                else:
                    errors.append("Failed to parse MySQL version")

        if errors:
            data['errors'] = errors
        return data

    def collect_data_os(self, profile, **kwargs):
        data = {}
        errors = []

        # release
        try:
            with open('/etc/os-release', 'rt') as f:
                output = f.read()
        except:
            errors.append("Failed to read /etc/release")
        else:
            release = {}
            pattern = re.compile(r'^([^=]+)=(.*)$')
            for line in output.strip().split('\n'):
                match = pattern.match(line)
                if match:
                    key, val = match.groups()
                    if val.startswith('"') and val.endswith('"'):
                        val = val.strip('"')
                    release[key] = val
            try:
                data['release_id'] = release['ID']
                data['release_version'] = release['VERSION_ID']
                data['release_full'] = release['PRETTY_NAME']
            except KeyError:
                errors.append("Failed to parse /etc/os-release")

        # timezone
        try:
            with open('/etc/timezone', 'rt') as f:
                output = f.read()
        except:
            errors.append("Failed to read /etc/timezone")
        else:
            data['timezone'] = output.strip()

        if errors:
            data['errors'] = errors
        return data

    def collect_data_php(self, profile, **kwargs):
        data = {}
        errors = []

        # release
        try:
            output = subprocess.check_output(['php', '--version'])
        except subprocess.CalledProcessError:
            errors.append("Failed to execute `php --version`")
        else:
            output = output.decode('utf_8').strip()
            data['release_full'] = output
            match = re.match(r'^PHP (\d+\.\d+\.\d+)-', output)
            if match:
                data['release_version'] = match.group(1)
            else:
                errors.append("Failed to parse PHP version")

        if errors:
            data['errors'] = errors
        return data

    def collect_data_postgresql(self, profile, **kwargs):
        data = {}
        errors = []

        # release
        try:
            output = subprocess.check_output(['psql', '--version'])
        except subprocess.CalledProcessError:
            errors.append("Failed to execute `psql --version`")
        else:
            output = output.decode('utf_8').strip()
            data['release_full'] = output
            match = re.match(r'^psql \(PostgreSQL\) (\d+\.\d+) ', output)
            if match:
                data['release_version'] = match.group(1)
            else:
                errors.append("Failed to parse PostgreSQL version")

        if errors:
            data['errors'] = errors
        return data

    def collect_data_python(self, profile, **kwargs):
        data = {}
        errors = []

        # envroot (required for others to work)
        envroot = profile._config_string('collect.python.envroot')
        if not envroot:
            errors.append("No envroot configured!")
        else:
            data['envroot'] = envroot

            # release
            python = os.path.join(envroot, 'bin/python')
            try:
                output = subprocess.check_output([python, '--version'])
            except subprocess.CalledProcessError:
                errors.append("Failed to execute `python --version`")
            else:
                output = output.decode('utf_8').strip()
                data['release_full'] = output
                match = re.match(r'^Python (\d+\.\d+\.\d+)', output)
                if match:
                    data['release_version'] = match.group(1)
                else:
                    errors.append("Failed to parse Python version")

        if errors:
            data['errors'] = errors
        return data

    def collect_data_rattail(self, profile, **kwargs):
        pass

    def normalize_errors(self, data, **kwargs):
        all_errors = []
        for key, value in data.items():
            if value:
                errors = value.pop('errors', None)
                if errors:
                    all_errors.extend(errors)
        if all_errors:
            data['errors'] = all_errors

    def submit_all_data(self, data, profile=None, **kwargs):
        profile = self.get_profile(profile)

        # TODO: any need for other types of submit logic?
        self.submit_data_tailbone_api(data, profile)

    def submit_data_tailbone_api(self, data, profile, **kwargs):
        api = TailboneAPIClient(self.config)

        data['uuid'] = profile.submit_uuid
        api.post(profile.submit_url, data=data)


class TelemetryProfile(ConfigProfile):
    """
    Holds configuration for a specific "profile" for use with
    telemetry.
    """
    section = 'rattail.telemetry'

    def load(self):

        keys = self._config_string('collect.keys')
        if not keys:
            keys = self._config_string('collect')
            if keys:
                warnings.warn(f"URGENT: instead of '{self.section}.{self.prefix}.collect', "
                              f"you should set '{self.section}.{self.prefix}.collect.keys'",
                              DeprecationWarning, stacklevel=2)
        self.collect_keys = self.config.parse_list(keys or 'os,postgresql,rattail,python')

        # nb. for now this is assumed to be `tailbone_api`
        #self.submit_type = self._config_string('submit')

        self.submit_url = self._config_string('submit.url')
        self.submit_uuid = self._config_string('submit.uuid')



# -*- coding: utf-8; -*-
"""grow label_profile.code

Revision ID: a7b7246e48e2
Revises: 95a66a59d420
Create Date: 2018-12-18 19:02:13.497715

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'a7b7246e48e2'
down_revision = '95a66a59d420'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # label_profile
    op.alter_column('label_profile', 'code', type_=sa.String(length=30))

    # label_batch
    op.alter_column('label_batch_row', 'label_code', type_=sa.String(length=30))


def downgrade():

    # label_batch
    op.alter_column('label_batch_row', 'label_code', type_=sa.String(length=3))

    # label_profile
    op.alter_column('label_profile', 'code', type_=sa.String(length=3))



# -*- coding: utf-8; -*-
"""grow product.item_id

Revision ID: aeea3c6fc6af
Revises: 43d766defb6a
Create Date: 2020-06-29 15:18:17.898191

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'aeea3c6fc6af'
down_revision = u'43d766defb6a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product
    op.alter_column('product', 'item_id', type_=sa.String(length=50))
    op.alter_column('product_version', 'item_id', type_=sa.String(length=50))


def downgrade():

    # product
    op.alter_column('product_version', 'item_id', type_=sa.String(length=20))
    op.alter_column('product', 'item_id', type_=sa.String(length=20))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Generator for 'rattail' projects
"""

import os

import colander


from rattail.projects import PoserProjectGenerator


class RattailProjectGenerator(PoserProjectGenerator):
    """
    Generator for "generic" Rattail projects
    """
    key = 'rattail'

    def make_schema(self, **kwargs):
        schema = super(RattailProjectGenerator, self).make_schema(**kwargs)

        # TODO: get rid of these after templates are updated

        schema.add(colander.SchemaNode(name='integrates_catapult',
                                       typ=colander.Boolean()))

        schema.add(colander.SchemaNode(name='integrates_corepos',
                                       typ=colander.Boolean()))

        schema.add(colander.SchemaNode(name='integrates_locsms',
                                       typ=colander.Boolean()))

        return schema

    def generate_project(self, output, context, **kwargs):
        super(RattailProjectGenerator, self).generate_project(
            output, context, **kwargs)

        package = os.path.join(output, context['pkg_name'])

        ##############################
        # fablib / machines
        ##############################

        if context['uses_fabric']:

            fablib = os.path.join(package, 'fablib')
            os.makedirs(fablib)

            self.generate('package/fablib/__init__.py.mako',
                          os.path.join(fablib, '__init__.py'),
                          context)

            self.generate('package/fablib/python.py.mako',
                          os.path.join(fablib, 'python.py'),
                          context)

            deploy = os.path.join(fablib, 'deploy')
            os.makedirs(deploy)

            python = os.path.join(deploy, 'python')
            os.makedirs(python)

            self.generate('package/fablib/deploy/python/premkvirtualenv.mako',
                          os.path.join(python, 'premkvirtualenv.mako'),
                          context)

            machines = os.path.join(output, 'machines')
            os.makedirs(machines)

            server = os.path.join(machines, 'server')
            os.makedirs(server)

            self.generate('machines/server/README.md.mako',
                          os.path.join(server, 'README.md'),
                          context)

            self.generate('machines/server/Vagrantfile.mako',
                          os.path.join(server, 'Vagrantfile'),
                          context)

            self.generate('machines/server/fabenv.py.dist.mako',
                          os.path.join(server, 'fabenv.py.dist'),
                          context)

            self.generate('machines/server/fabric.yaml.dist',
                          os.path.join(server, 'fabric.yaml.dist'))

            self.generate('machines/server/fabfile.py.mako',
                          os.path.join(server, 'fabfile.py'),
                          context)

            deploy = os.path.join(server, 'deploy')
            os.makedirs(deploy)

            poser = os.path.join(deploy, context['folder'])
            os.makedirs(poser)

            if context['integrates_catapult']:
                self.generate('machines/server/deploy/poser/freetds.conf.mako_',
                              os.path.join(poser, 'freetds.conf.mako'))
                self.generate('machines/server/deploy/poser/odbc.ini',
                              os.path.join(poser, 'odbc.ini'))

            self.generate('machines/server/deploy/poser/rattail.conf.mako',
                          os.path.join(poser, 'rattail.conf.mako'),
                          context)

            self.generate('machines/server/deploy/poser/cron.conf.mako',
                          os.path.join(poser, 'cron.conf'),
                          context)

            self.generate('machines/server/deploy/poser/web.conf.mako',
                          os.path.join(poser, 'web.conf.mako'),
                          context)

            self.generate('machines/server/deploy/poser/supervisor.conf.mako',
                          os.path.join(poser, 'supervisor.conf'),
                          context)

            self.generate('machines/server/deploy/poser/overnight.sh.mako',
                          os.path.join(poser, 'overnight.sh'),
                          context)

            self.generate('machines/server/deploy/poser/overnight-wrapper.sh.mako',
                          os.path.join(poser, 'overnight-wrapper.sh'),
                          context)

            self.generate('machines/server/deploy/poser/crontab.mako',
                          os.path.join(poser, 'crontab.mako'),
                          context)

            self.generate('machines/server/deploy/poser/upgrade.sh.mako',
                          os.path.join(poser, 'upgrade.sh'),
                          context)

            self.generate('machines/server/deploy/poser/tasks.py.mako',
                          os.path.join(poser, 'tasks.py'),
                          context)

            self.generate('machines/server/deploy/poser/upgrade-wrapper.sh.mako',
                          os.path.join(poser, 'upgrade-wrapper.sh'),
                          context)

            self.generate('machines/server/deploy/poser/sudoers.mako',
                          os.path.join(poser, 'sudoers'),
                          context)

            self.generate('machines/server/deploy/poser/logrotate.conf.mako',
                          os.path.join(poser, 'logrotate.conf'),
                          context)



# -*- coding: utf-8 -*-
"""add 'missing' amounts for purchase

Revision ID: 3f3523703313
Revises: 24577b3bda93
Create Date: 2021-12-12 15:06:09.873087

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '3f3523703313'
down_revision = '24577b3bda93'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_item
    op.add_column('purchase_item', sa.Column('cases_missing', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_item', sa.Column('units_missing', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_item', sa.Column('po_case_size', sa.Numeric(precision=8, scale=2), nullable=True))
    op.add_column('purchase_item', sa.Column('invoice_case_size', sa.Numeric(precision=8, scale=2), nullable=True))

    # batch_purchase_row
    op.add_column('batch_purchase_row', sa.Column('cases_missing', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('batch_purchase_row', sa.Column('units_missing', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('batch_purchase_row', sa.Column('po_case_size', sa.Numeric(precision=8, scale=2), nullable=True))
    op.add_column('batch_purchase_row', sa.Column('invoice_case_size', sa.Numeric(precision=8, scale=2), nullable=True))


def downgrade():

    # batch_purchase_row
    op.drop_column('batch_purchase_row', 'units_missing')
    op.drop_column('batch_purchase_row', 'cases_missing')
    op.drop_column('batch_purchase_row', 'invoice_case_size')
    op.drop_column('batch_purchase_row', 'po_case_size')

    # purchase_item
    op.drop_column('purchase_item', 'units_missing')
    op.drop_column('purchase_item', 'cases_missing')
    op.drop_column('purchase_item', 'invoice_case_size')
    op.drop_column('purchase_item', 'po_case_size')



# -*- coding: utf-8 -*-
"""grow vendorcatalog.cost

Revision ID: a809caf23cf0
Revises: 9de6ffc1d2af
Create Date: 2017-12-21 20:11:02.812699

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'a809caf23cf0'
down_revision = u'9de6ffc1d2af'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # vendor_catalog_row
    op.alter_column('vendor_catalog_row', 'case_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('vendor_catalog_row', 'unit_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('vendor_catalog_row', 'old_case_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('vendor_catalog_row', 'old_unit_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('vendor_catalog_row', 'case_cost_diff', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('vendor_catalog_row', 'unit_cost_diff', type_=sa.Numeric(precision=10, scale=5))


def downgrade():

    # vendor_catalog_row
    op.alter_column('vendor_catalog_row', 'unit_cost_diff', type_=sa.Numeric(precision=9, scale=5))
    op.alter_column('vendor_catalog_row', 'case_cost_diff', type_=sa.Numeric(precision=9, scale=5))
    op.alter_column('vendor_catalog_row', 'old_unit_cost', type_=sa.Numeric(precision=9, scale=5))
    op.alter_column('vendor_catalog_row', 'old_case_cost', type_=sa.Numeric(precision=9, scale=5))
    op.alter_column('vendor_catalog_row', 'unit_cost', type_=sa.Numeric(precision=9, scale=5))
    op.alter_column('vendor_catalog_row', 'case_cost', type_=sa.Numeric(precision=9, scale=5))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Org Handler
"""

from sqlalchemy import orm

from rattail.app import GenericHandler


class OrgHandler(GenericHandler):
    """
    Base class and default implementation for org handlers.

    This is meant to provide logic around the "organization" - for
    instance stores and departments.
    """

    def get_store(self, session, key):
        """
        Locate and return a store for the given key, if possible.

        First the key is assumed to be a ``Store.id`` value.  If no
        matches are found, then it looks for a special setting in the
        database.  If one is found, ``get_store()`` is called again
        with its value.

        :param session: Active database session.

        :param key: Value to use when searching for the store.

        :returns: The :class:`~rattail.db.model.Store` instance if
           found; otherwise ``None``.
        """
        model = self.model

        # Store.uuid match?
        store = session.get(model.Store, key)
        if store:
            return store

        # Store.id match?
        try:
            return session.query(model.Store).filter_by(id=key).one()
        except orm.exc.NoResultFound:
            pass

        # try settings, if value then recurse.
        key = self.app.get_setting(session, f'rattail.store.{key}')
        if key:
            return self.get_store(session, key)

    def get_department(self, session, key):
        """
        Locate and return a department for the given key, if possible.

        First the key is assumed to be a ``Department.number`` value.
        If no matches are found, then it looks for a special setting
        in the database.  If one is found, ``get_department()`` is
        called again with its value.

        :param session: Active database session.

        :param key: Value to use when searching for the department.

        :returns: The :class:`~rattail.db.model.Department` instance
           if found; otherwise ``None``.
        """
        model = self.model

        # Department.uuid match?
        department = session.get(model.Department, str(key))
        if department:
            return department

        # Department.number match?
        if isinstance(key, int) or key.isdigit():
            try:
                return session.query(model.Department).filter_by(number=key).one()
            except orm.exc.NoResultFound:
                pass

        # try settings, if value then recurse.
        key = self.app.get_setting(session, f'rattail.department.{key}')
        if key is not None:
            return self.get_department(session, key)

    def get_subdepartment(session, key):
        """
        Locate and return a subdepartment for the given key, if
        possible.

        First the key is assumed to be a ``Subdepartment.number``
        value.  If no matches are found, then it looks for a special
        setting in the database.  If one is found,
        ``get_subdepartment()`` is called again with its value.

        :param session: Active database session.

        :param key: Value to use when searching for the subdepartment.

        :returns: The :class:`~rattail.db.model.Subdepartment`
           instance if found; otherwise ``None``.
        """
        model = self.model

        # Subdepartment.uuid match?
        subdepartment = session.get(model.Subdepartment, key)
        if subdepartment:
            return subdepartment

        # Subdepartment.number match?
        if isinstance(key, int) or key.isdigit():
            try:
                return session.query(model.Subdepartment).filter_by(number=key).one()
            except orm.exc.NoResultFound:
                pass

        # try settings, if value then recurse.
        key = self.app.get_setting(session, f'rattail.subdepartment.{key}')
        if key is not None:
            return self.get_subdepartment(session, key)



# -*- coding: utf-8 -*-
"""add Email.invalid flag

Revision ID: 4da3fe4a368f
Revises: 14d23631733d
Create Date: 2015-12-17 11:10:56.104030

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '4da3fe4a368f'
down_revision = u'14d23631733d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # email
    op.add_column('email', sa.Column('invalid', sa.Boolean(), nullable=True))


def downgrade():

    # email
    op.drop_column('email', 'invalid')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Problem Reports for mailmon
"""

from __future__ import unicode_literals, absolute_import

import six

from rattail.problems import RattailProblemReport
from rattail.mailmon import MailmonConnection
from rattail.mailmon.config import load_mailmon_profiles


class MailmonMisses(RattailProblemReport):
    """
    Looks for any mail in IMAP accounts which appears to have been
    missed by the mailmon daemon.
    """
    problem_key = 'mailmon_misses'
    problem_title = "Mailmon misses"

    def find_problems(self, **kwargs):
        problems = []

        # loop thru all accounts
        monitored = load_mailmon_profiles(self.config)
        for account in six.itervalues(monitored['__accounts__']):
            cxn = MailmonConnection(self.config, account)
            cxn.connect()

            # loop thru all profiles for the account
            for profile in account.monitored:
                count = cxn.select_folder(profile.imap_folder)
                if count:
                    problems.append((account, profile, count))

            cxn.disconnect()

        return problems



# -*- coding: utf-8; -*-
"""add work orders

Revision ID: 62cfde95d655
Revises: 9c111f4b5451
Create Date: 2022-08-09 19:43:01.326910

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '62cfde95d655'
down_revision = '9c111f4b5451'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types


workorder_id_seq = sa.Sequence('workorder_id_seq')


def upgrade():
    from sqlalchemy.schema import CreateSequence

    # id sequence
    op.execute(CreateSequence(workorder_id_seq))

    # workorder
    op.create_table('workorder',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('customer_uuid', sa.String(length=32), nullable=False),
                    sa.Column('date_submitted', sa.Date(), nullable=True),
                    sa.Column('date_received', sa.Date(), nullable=True),
                    sa.Column('date_released', sa.Date(), nullable=True),
                    sa.Column('date_delivered', sa.Date(), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.Column('status_code', sa.Integer(), nullable=False),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.ForeignKeyConstraint(['customer_uuid'], ['customer.uuid'], name='workorder_fk_customer'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # workorder_event
    op.create_table('workorder_event',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('workorder_uuid', sa.String(length=32), nullable=False),
                    sa.Column('type_code', sa.Integer(), nullable=False),
                    sa.Column('occurred', sa.DateTime(), nullable=False),
                    sa.Column('user_uuid', sa.String(length=32), nullable=False),
                    sa.Column('note', sa.Text(), nullable=True),
                    sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'], name='workorder_event_fk_user'),
                    sa.ForeignKeyConstraint(['workorder_uuid'], ['workorder.uuid'], name='workorder_event_fk_workorder'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():
    from sqlalchemy.schema import DropSequence

    # workorder_event
    op.drop_table('workorder_event')

    # workorder
    op.drop_table('workorder')

    # id sequence
    op.execute(DropSequence(workorder_id_seq))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for Generating Features
"""

from __future__ import unicode_literals, absolute_import

import six

import colander

from rattail.app import GenericHandler
from rattail.util import load_entry_points


class FeatureHandler(GenericHandler):
    """
    Base class for feature handlers.
    """
    entry_point_section = 'rattail.features'

    def all_features(self):
        """
        Returns a dict of available features, which are registered via
        setuptools entry points.
        """
        return load_entry_points(self.entry_point_section)

    def all_feature_types(self):
        """
        Returns a list of available feature type keys.
        """
        return list(self.all_features())

    def iter_features(self):
        """
        Iterate over all features.
        """
        for factory in six.itervalues(self.all_features()):
            yield factory(self.config)

    def get_feature(self, key):
        """
        Returns the specific feature identified by type key.
        """
        features = self.all_features()
        if key in features:
            return features[key](self.config)

    def make_schema(self, **kwargs):

        class Schema(colander.MappingSchema):

            app_prefix = colander.SchemaNode(colander.String())

            app_cap_prefix = colander.SchemaNode(colander.String())

        return Schema(**kwargs)

    def get_defaults(self):
        return {
            'app_prefix': self.config.app_title().lower(),
            'app_cap_prefix': self.config.app_title(),
        }

    def do_generate(self, feature, **kwargs):
        """
        Generate code and instructions for new feature.
        """
        return feature.generate(**kwargs)



# -*- coding: utf-8; -*-
"""add receiving.truck_dump_children_first

Revision ID: 40a4855cb22d
Revises: 4db5dbf725f5
Create Date: 2019-02-05 11:37:03.807392

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '40a4855cb22d'
down_revision = '4db5dbf725f5'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch
    op.add_column('purchase_batch', sa.Column('truck_dump_children_first', sa.Boolean(), nullable=True))
    op.add_column('purchase_batch', sa.Column('truck_dump_ready', sa.Boolean(), nullable=True))


def downgrade():

    # purchase_batch
    op.drop_column('purchase_batch', 'truck_dump_ready')
    op.drop_column('purchase_batch', 'truck_dump_children_first')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Report Commands
"""

import logging

from sqlalchemy import orm

from rattail.commands import Subcommand
from rattail.time import get_sunday, get_monday


log = logging.getLogger(__name__)


class ReportSubcommand(Subcommand):
    """
    Generate and email a report
    """
    report_key = None
    email_key = None
    simple_report_name = "(Unnamed)"

    def ensure_report_key(self):
        if not self.report_key:
            raise RuntimeError("must define {}.report_key".format(
                self.__class__.__name__))


class WeeklyReport(ReportSubcommand):
    """
    Generate and email a weekly report
    """

    def run(self, args):
        self.ensure_report_key()
        session = self.make_session()
        model = self.model

        # first must determine most recent complete Mon-Sun date range
        # TODO: should probably be more flexible about date range..
        today = self.app.today()
        sunday = get_sunday(today)
        monday = get_monday(sunday)
        start_date = monday
        end_date = sunday

        # determine naming for the report
        report_name = "{} {} thru {}".format(
            self.simple_report_name,
            start_date.strftime("%Y-%m-%d"),
            end_date.strftime("%Y-%m-%d"))

        try:
            # see if this report has already been ran
            output = session.query(model.ReportOutput)\
                            .filter(model.ReportOutput.report_type == self.report_key)\
                            .filter(model.ReportOutput.report_name == report_name)\
                            .one()

        except orm.exc.NoResultFound:

            # generate report file and commit result
            handler = self.app.get_report_handler()
            report = handler.get_report(self.report_key)
            params = {'start_date': start_date, 'end_date': end_date}
            user = self.get_runas_user(session=session)
            output = handler.generate_output(session, report, params, user,
                                             progress=self.progress)
            session.commit()

            # try to make url for report
            report_url = None
            base_url = self.config.base_url()
            if base_url:
                report_url = '{}/reports/generated/{}'.format(
                    base_url, output.uuid)

            # send report output as email
            email_key = self.email_key or self.report_key
            handler.email_output(report, output, email_key,
                                 extra_data={'report_url': report_url})

        else:
            log.warning("report output already exists: %s", report_name)

        session.close()



# -*- coding: utf-8 -*-
"""add purchase_batch_row_claim

Revision ID: b5aa18867ab3
Revises: 977a76f259bd
Create Date: 2018-05-17 11:43:14.639019

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'b5aa18867ab3'
down_revision = u'977a76f259bd'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch_row_claim
    op.create_table('purchase_batch_row_claim',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('claiming_row_uuid', sa.String(length=32), nullable=False),
                    sa.Column('claimed_row_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cases_received', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('units_received', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('cases_damaged', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('units_damaged', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('cases_expired', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('units_expired', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.ForeignKeyConstraint(['claimed_row_uuid'], [u'purchase_batch_row.uuid'], name=u'purchase_batch_row_claim_fk_claimed_row'),
                    sa.ForeignKeyConstraint(['claiming_row_uuid'], [u'purchase_batch_row.uuid'], name=u'purchase_batch_row_claim_fk_claiming_row'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # purchase_batch_row_claim
    op.drop_table('purchase_batch_row_claim')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Application Configuration
"""

import importlib
import os
import re
import sys
import shlex
import datetime
import configparser
import warnings
import logging
import logging.config

import config as configuration

from rattail.util import load_entry_points, load_object
from rattail.exceptions import WindowsExtensionsNotInstalled, ConfigurationError
from rattail.files import temp_path
from rattail.logging import TimeConverter


log = logging.getLogger(__name__)


def parse_bool(value):
    """
    Derive a boolean from the given string value.
    """
    if value is None:
        return None
    if isinstance(value, bool):
        return value
    if str(value).lower() in ('true', 'yes', 'y', 'on', '1'):
        return True
    return False


def parse_list(value):
    """
    Parse a configuration value, splitting by whitespace and/or commas and
    taking quoting into account, etc., yielding a list of strings.
    """
    if value is None:
        return []
    parser = shlex.shlex(value)
    parser.whitespace += ','
    parser.whitespace_split = True
    values = list(parser)
    for i, value in enumerate(values):
        if value.startswith('"') and value.endswith('"'):
            values[i] = value[1:-1]
        elif value.startswith("'") and value.endswith("'"):
            values[i] = value[1:-1]
    return values


class RattailConfigWrapper:
    """
    Simple wrapper around *either* the legacy :class:`RattailConfig`
    or the newer :class:`RattailConfiguration` object.

    .. note::

       This class will go away once migration to the newer
       ``RattailConfiguration`` is complete.
    """

    def __init__(self, *args, **kwargs):

        # first make legacy config
        self.__dict__['config'] = RattailConfig(*args, **kwargs)

        # but use new-style instead, if config says so
        if self.config.getbool('rattail.config', 'use_configuration', usedb=False):
            self.__dict__['config'] = RattailConfiguration(*args, **kwargs)
            self.__dict__['_style'] = 'configuration'
            self.config.configure_logging()

        else: # legacy config
            self.__dict__['_style'] = 'legacy'
            self.config.configure_logging()

    def __getattr__(self, name):
        return getattr(self.config, name)

    def __setattr__(self, name, value):
        setattr(self.config, name, value)

    def __repr__(self):
        return f"RattailConfigWrapper(style={self._style})"


class RattailConfigMixin:
    """
    Extra methods for Rattail config classes.
    """

    def get_app(self):
        """
        Returns the global :class:`~rattail.app.AppHandler` instance,
        creating it if necessary.
        """
        if not hasattr(self, 'app'):
            spec = self.get('rattail', 'app.handler', usedb=False,
                            default='rattail.app:AppHandler')
            factory = load_object(spec)
            self.app = factory(self)
        return self.app

    def node_type(self, default=None):
        """
        Returns the "type" of current node.  What this means will
        generally depend on the app logic.
        """
        try:
            return self.require('rattail', 'node_type', usedb=False)
        except ConfigurationError:
            if default:
                return default
            raise

    def production(self):
        """
        Returns boolean indicating whether the app is running in
        production mode
        """
        return self.getbool('rattail', 'production', default=False)

    def get_model(self):
        """
        Returns a reference to configured 'model' module; defaults to
        :mod:`rattail.db.model`.
        """
        spec = self.get('rattail', 'model', usedb=False,
                        default='rattail.db.model')
        return importlib.import_module(spec)

    def get_enum(self, **kwargs):
        """
        Returns a reference to configured "enum" module; defaults to
        :mod:`rattail.enum`.
        """
        kwargs.setdefault('usedb', False)
        spec = self.get('rattail', 'enum', default='rattail.enum', **kwargs)
        return importlib.import_module(spec)

    def get_trainwreck_model(self):
        """
        Returns a reference to the configured data 'model' module for
        Trainwreck.  Note that there is *not* a default value for
        this; it must be configured.
        """
        spec = self.require('rattail.trainwreck', 'model', usedb=False)
        return importlib.import_module(spec)

    def versioning_enabled(self):
        """
        Returns boolean indicating whether data versioning is enabled.
        """
        return self.getbool('rattail.db', 'versioning.enabled', usedb=False,
                            default=False)

    def product_key(self, **kwargs):
        """
        Deprecated; instead please see
        :meth:`rattail.app.AppHandler.get_product_key_field()`.
        """
        warnings.warn("config.product_key() is deprecated; please "
                      "use app.get_product_key_field() instead",
                      DeprecationWarning, stacklevel=2)
        return self.get_app().get_product_key_field()

    def product_key_title(self, key=None):
        """
        Deprecated; instead please see
        :meth:`rattail.app.AppHandler.get_product_key_label()`.
        """
        warnings.warn("config.product_key_title() is deprecated; please "
                      "use app.get_product_key_label() instead",
                      DeprecationWarning, stacklevel=2)
        return self.get_app().get_product_key_label(field=key)

    def app_package(self, default=None):
        """
        Returns the name of Python package for the top-level app.
        """
        if not default:
            return self.require('rattail', 'app_package')
        return self.get('rattail', 'app_package', default=default)

    def app_title(self, **kwargs):
        """ DEPRECATED """
        # TODO: should put a deprecation warning here, but it could
        # make things noisy for a while and i'm not ready for that
        app = self.get_app()
        return app.get_title(**kwargs)

    def node_title(self, **kwargs):
        """ DEPRECATED """
        # TODO: should put a deprecation warning here, but it could
        # make things noisy for a while and i'm not ready for that
        app = self.get_app()
        return app.get_node_title(**kwargs)

    def running_from_source(self):
        """
        Returns boolean indicating whether the app is running from
        source, as opposed to official release.
        """
        return self.getbool('rattail', 'running_from_source', default=False)

    def demo(self):
        """
        Returns boolean indicating whether the app is running in demo mode
        """
        return self.getbool('rattail', 'demo', default=False)

    def appdir(self, require=True, **kwargs):
        """
        Returns path to the 'app' dir, if known.
        """
        if require:
            path = os.path.join(sys.prefix, 'app')
            kwargs.setdefault('default', path)
        kwargs.setdefault('usedb', False)
        return self.get('rattail', 'appdir', **kwargs)

    def datadir(self, require=True):
        """
        Returns path to the 'data' dir, if known.
        """
        get = self.require if require else self.get
        return get('rattail', 'datadir')

    def workdir(self, require=True):
        """
        Returns path to the 'work' dir, if known.
        """
        get = self.require if require else self.get
        return get('rattail', 'workdir')

    def batch_filedir(self, key=None):
        """
        Returns path to root folder where batches (optionally of type
        'key') are stored.
        """
        path = os.path.abspath(self.require('rattail', 'batch.files'))
        if key:
            return os.path.join(path, key)
        return path

    def batch_filepath(self, key, uuid, filename=None, makedirs=False):
        """
        Returns absolute path to a batch's data folder, with optional
        filename appended.  If ``makedirs`` is set, the batch data
        folder will be created if it does not already exist.
        """
        rootdir = self.batch_filedir(key)
        filedir = os.path.join(rootdir, uuid[:2], uuid[2:])
        if makedirs and not os.path.exists(filedir):
            os.makedirs(filedir)
        if filename:
            return os.path.join(filedir, filename)
        return filedir

    def export_filedir(self, key=None):
        """
        Returns path to root folder where exports (optionally of type
        'key') are stored.
        """
        path = self.get('rattail', 'export.files')
        if not path:
            path = os.path.join(self.appdir(), 'data', 'exports')
        path = os.path.abspath(path)
        if key:
            return os.path.join(path, key)
        return path

    def export_filepath(self, key, uuid, filename=None, makedirs=False):
        """
        Returns absolute path to export data file, generated from the given args.
        """
        rootdir = self.export_filedir(key)
        filedir = os.path.join(rootdir, uuid[:2], uuid[2:])
        if makedirs and not os.path.exists(filedir):
            os.makedirs(filedir)
        if filename:
            return os.path.join(filedir, filename)
        return filedir

    def upgrade_filedir(self):
        """
        Returns path to root folder where upgrade files are stored.
        """
        path = os.path.abspath(self.require('rattail.upgrades', 'files'))
        return path

    def upgrade_filepath(self, uuid, filename=None, makedirs=False):
        """
        Returns absolute path to upgrade data file, generated from the given args.
        """
        rootdir = self.upgrade_filedir()
        filedir = os.path.join(rootdir, uuid[:2], uuid[2:])
        if makedirs and not os.path.exists(filedir):
            os.makedirs(filedir)
        if filename:
            return os.path.join(filedir, filename)
        return filedir

    def upgrade_command(self, default='/bin/sleep 30'):
        """
        Returns command to be used when performing upgrades.
        """
        # TODO: what were those reasons then..?
        # NOTE: we don't allow command to be specified in DB, for
        # security reasons..
        return self.getlist('rattail.upgrades', 'command', usedb=False,
                            default=default)

    def base_url(self):
        """
        Returns the configured "base" (root) URL for the web app.
        """
        # first try "generic" config option
        url = self.get('rattail', 'base_url')

        # or use tailbone as fallback, since it's most likely
        if url is None:
            url = self.get('tailbone', 'url')

        if url is not None:
            return url.rstrip('/')

    def datasync_url(self, **kwargs):
        """
        Returns configured URL for managing datasync daemon.
        """
        return self.get('rattail.datasync', 'url', **kwargs)

    def single_store(self):
        """
        Returns boolean indicating whether the system is configured to behave
        as if it belongs to a single Store.
        """
        return self.getbool('rattail', 'single_store', default=False)

    def get_store(self, session):
        """
        Returns a :class:`rattail.db.model.Store` instance
        corresponding to app config, or ``None``.
        """
        store = self.get('rattail', 'store')
        if store:
            app = self.get_app()
            org_handler = app.get_org_handler()
            return org_handler.get_store(session, store)


class RattailConfiguration(RattailConfigMixin):
    """
    Configuration for Rattail apps.

    .. warning::

       I'm documenting this as though it's complete, but as of writing
       this class is **not used** unless config says to, via:

       .. code-block:: ini

          [rattail.config]
          use_configuration = true

    A single instance of this class is created on app startup, by way
    of calling :func:`make_config()`.

    Even though this config object is "global" in nature, it must
    be passed around to all areas of the app; you cannot simply
    import it.  In other words::

       # this won't work!
       from rattail.config import global_config_object

    The global config object is mainly responsible for providing
    config values to the app, via :meth:`get()` and similar methods.

    The config object has more than one place to look when finding
    values.  This can vary somewhat but usually the priority for
    lookup is like:

    * settings table in the DB
    * one or more INI files
    * "defaults" provided by app logic

    .. attribute:: configuration

       Reference to the ``ConfigurationSet`` instance which houses the
       full set of config values which are kept in memory.  This does
       *not* contain settings from DB, but *does* contain
       :attr:`defaults` as well as values read from INI files.

    .. attribute:: defaults

       Reference to the ``Configuration`` instance containing config
       *default* values.  This is exposed in case it's useful, but in
       practice you should not update it directly; instead use
       :meth:`setdefault()`.

    .. attribute:: files_read

       List of all INI config files which were read on app startup.
       These are listed in the same order as they were read.  This
       sequence also reflects priority for value lookups, i.e. the
       first file with the value wins.

    .. attribute:: usedb

       Flag indicating whether values should ever be looked up from
       the settings table in DB.  Note that you can override this when
       calling :meth:`get()`.

    .. attribute:: preferdb

       Flag indicating whether DB settings should be preferred over
       the values from INI files or app defaults.  Note that you can
       override this when calling :meth:`get()`.

    .. attribute:: versioning_has_been_enabled

       Flag indicating whether SQLAlchemy-Continuum versioning has
       been enabled for the running app.  This gets set when
       :func:`~rattail.db.config.configure_versioning()` happens.
    """

    def __init__(
            self,
            files=[],
            usedb=None,
            preferdb=None,
    ):

        self.files_read = []
        configs = []

        # read all files requested
        for path in files:
            self._load_ini_configs(path, configs, require=True)

        # add 'defaults' config, for use w/ setdefault()
        self.defaults = configuration.Configuration({})
        configs.append(self.defaults)

        # master config set
        self.configuration = configuration.ConfigurationSet(*configs)

        # usedb flag
        self.usedb = usedb
        if self.usedb is None:
            self.usedb = self.get_bool('rattail.config.usedb',
                                       usedb=False, default=False)

        # preferdb flag
        self.preferdb = preferdb
        if self.usedb and self.preferdb is None:
            self.preferdb = self.get_bool('rattail.config.preferdb',
                                          usedb=False, default=False)

        # attempt to detect lack of sqlalchemy libraries etc.  this
        # allows us to avoid installing those on a machine which will
        # not need to access a database etc.
        if self.usedb:
            try:
                from rattail.db import Session
            except ImportError: # pragma: no cover
                log.warning("config created with `usedb = True`, but can't import "
                            "`rattail.db.Session`, so setting `usedb = False` instead",
                            exc_info=True)
                self.usedeb = False
                self.preferdb = False

        # this is always false, unless/until it becomes true
        self.versioning_has_been_enabled = False

    @property
    def prioritized_files(self):
        return self.files_read

    def _load_ini_configs(self, path, configs, require=True):
        path = os.path.abspath(path)

        # try to load config from the given path
        try:
            config = configuration.config_from_ini(path, read_from_file=True)
        except FileNotFoundError:
            if not require:
                log.warning("INI config file not found: %s", path)
                return
            raise

        # ok add that one to the mix
        configs.append(config)
        self.files_read.append(path)

        # need parent folder of that path, for %(here)s interpolation
        here = os.path.dirname(path)

        # bring in any "required" files
        requires = config.get('rattail.config.require')
        if requires:
            for path in parse_list(requires):
                path = path % {'here': here}
                self._load_ini_configs(path, configs, require=True)

        # bring in any "included" files
        includes = config.get('rattail.config.include')
        if includes:
            for path in parse_list(includes):
                path = path % {'here': here}
                self._load_ini_configs(path, configs, require=False)

    def setdefault(self, *args):
        """
        Establish a default config value.

        Positional arguments to this method may vary.  The *new*
        (optional) way is to pass just two args: ``key`` and
        ``value``.  But the traditional way was to pass *three* args:
        ``section``, ``option`` and ``value``.

        The main point of this is to provide a default value when none
        is actually configured, i.e. via INI file or DB settings.

        Note that if the app already has a default value for the given
        key, this method will *not* replace that value.  So the first
        call with given key will set default value, then subsequent
        calls for that key have no effect.

        :returns: The current config value, *outside of the DB*.  For
           various reasons this method may not be able to lookup
           settings from the DB, e.g. during app init.  So it can only
           determine the value per INI files + config defaults.
        """
        # figure out what sort of args were passed
        if len(args) == 2:
            key, value = args
        elif len(args) == 3:
            section, option, value = args
            key = f'{section}.{option}'
        else:
            raise ValueError("must pass either 2 args (key, value), "
                             "or 3 args (section, option, value)")

        # set default value, if not already set
        self.defaults.setdefault(key, value)

        # get current value, sans db
        return self.get(key, usedb=False)

    def get(self, *args, **kwargs):
        """
        Retrieve a value from config.

        Positional arguments to this method may vary.  The *new*
        (optional) way is to pass just one ``key`` arg.  But the
        traditional way was to pass *two* args, ``section`` and
        ``option``.

        Any keyword arguments passed, if applicable, will have
        identical meanings as for :meth:`RattailConfig.get()`.

        :returns: Value as string (always!).
        """
        # figure out what sort of args were passed
        if len(args) == 1:
            key = args[0]
        elif len(args) == 2:
            section, option = args
            key = f'{section}.{option}'
        else:
            raise ValueError("must pass either 1 arg (key), "
                             "or 2 args (section, option)")

        if kwargs.get('require') and 'default' in kwargs:
            raise ValueError("must not specify default value when require=True")

        # caller may specify (part of) error message, if applicable.
        # but we remove it from kwargs for sake of other calls below
        msg = kwargs.pop('msg', None)

        # should we use db?
        usedb = kwargs.get('usedb')
        usedb = usedb if usedb is not None else self.usedb

        # should we prefer db?
        preferdb = False
        if usedb:
            preferdb = kwargs.get('preferdb')
            if preferdb is None:
                preferdb = self.preferdb

        # read from db first if so requested
        if usedb and preferdb:
            value = self.get_from_db(key, session=kwargs.get('session'))
            if value is not None:
                return value

        # read from defaults + INI files
        value = self.configuration.get(key)
        if value is not None:
            # TODO: in most cases the distinction of "key"
            # vs. "section, option" is transparent, but in some cases
            # not so.  e.g. this came up when fetching the value for
            # key of 'tailbone.menus' which was now ambiguous..
            #
            # e.g. older code uses config like:
            #
            # [tailbone]
            # menus = poser.web.menus
            #
            # but newer code looks elsewhere:
            #
            # [tailbone.menus]
            # handler = poser.web.menus:PoserMenuHandler
            #
            # with the above, config.get('tailbone.menus') will never
            # return 'poser.web.menus' and instead will always return
            # a "sub" config object with 'handler' key!
            #
            # this also came up for datasync watcher/consumer config,
            # where some keys became ambiguous in similar way.
            #
            # so now any such keys which would be ambiguous, must
            # simply be abandoned by all app logic, and unambiguous
            # keys must be used instead.  this requires updating logic
            # as well as config files and/or DB settings, and with a
            # lengthy grace period with fallback to old setting name.
            # note however the fallback will *not* work with this new
            # RattailConfiguration class, only the older RattailConfig!
            if not isinstance(value, configuration.Configuration):
                return value
            log.warning("ambiguous config key '%s' returns: %s", key, value)
            warnings.warn(f"ambiguous config key '{key}' returns: {value}",
                          DeprecationWarning, stacklevel=2)

        # read from db last if so requested
        if usedb and not preferdb:
            value = self.get_from_db(key, session=kwargs.get('session'))
            if value is not None:
                return value

        # raise error if required value not found
        if kwargs.get('require'):
            msg = (f"{msg or 'missing or invalid config'}; "
                   f"please set config value for: {key}")
            raise ConfigurationError(msg)

        # give the default value if specified
        if 'default' in kwargs:
            return kwargs['default']

    def require(self, *args, **kwargs):
        """
        Retrieve a value from config, or raise error if no value can
        be found.  This is just a shortcut, so these work the same::

           config.get('foo', require=True)

           config.require('foo')
        """
        kwargs['require'] = True
        return self.get(*args, **kwargs)

    def get_from_db(self, key, session=None):
        """
        Retrieve a config value from database settings table.
        """
        app = self.get_app()
        with app.short_session(session=session) as s:
            return app.get_setting(s, key)

    def get_bool(self, *args, **kwargs):
        """
        Retrieve a boolean value from config.
        """
        value = self.get(*args, **kwargs)
        return self.parse_bool(value)

    getbool = get_bool

    def parse_bool(self, value):
        """
        Coerce the given value to a boolean.
        """
        return parse_bool(value)

    def get_int(self, *args, **kwargs):
        """
        Retrieve an integer value from config.
        """
        value = self.get(*args, **kwargs)
        if value is None:
            return None
        if value == '' and 'default' in kwargs:
            return kwargs['default']
        if isinstance(value, int):
            return value
        return int(value)

    getint = get_int

    def get_date(self, *args, **kwargs):
        """
        Retrieve a date value from config.
        """
        value = self.get(*args, **kwargs)
        app = self.get_app()
        return app.parse_date(value)

    getdate = get_date

    def get_list(self, *args, **kwargs):
        """
        Retrieve a list of string values from a single config value.
        """
        value = self.get(*args, **kwargs)
        if value is None:
            return None
        if isinstance(value, str):
            return self.parse_list(value)
        return value            # maybe a caller-provided default?

    getlist = get_list

    def parse_list(self, value):
        """
        Coerce the given value to a list.
        """
        return parse_list(value)

    def make_list_string(self, values):
        """
        Coerce the given list of values to a string, for config
        storage.  If this string is later parsed via
        :meth:`parse_list()` then it should return the same list of
        values.
        """
        final = []
        for value in values:
            if ' ' in value:
                quote = '"' if "'" in value else "'"
                value = f"{quote}{value}{quote}"
            final.append(value)
        return ', '.join(final)

    def get_dict(self, prefix):
        """
        Retrieve a particular group of values, as a dictionary.

        Please note, this will only return values from INI files +
        defaults.  It will *not* return values from DB settings.

        The ``prefix`` is analogous to the ``section`` arg used in
        :meth:`RattailConfig.get_dict()`.  However in this case it is
        not *necessarily* a traditional "section" name from INI file.
        Rather, it refers to "any" logical group under which other
        config values may be found.
        """
        try:
            values = self.configuration[prefix]
        except KeyError:
            return {}

        return values.as_dict()

    def configure_logging(self):
        """
        This first checks current config to determine whether or not we're
        supposed to be configuring logging at all.  If not, nothing more is
        done.

        If we are to configure logging, then this will save the current config
        parser defaults to a temporary file, and use this file to configure
        Python's standard logging module.
        """
        if not self.get_bool('rattail.config.configure_logging', usedb=False):
            return

        # coerce all logged timestamps to the local timezone, if possible
        timezone = self.get('rattail', 'timezone.default', usedb=False)
        if timezone:
            logging.Formatter.converter = TimeConverter(self)

        # write current values to file suitable for logging auto-config
        path = self._write_logging_config_file()
        try:
            logging.config.fileConfig(path, disable_existing_loggers=False)
        except configparser.NoSectionError as error:
            log.warning("tried to configure logging, but got NoSectionError: %s",
                        error)
        else:
            log.debug("configured logging")
        finally:
            os.remove(path)

        # if we could not coerce timestamps, log a warning about that
        if not timezone:
            log.warning("timezone is not set, so logging timestamps may "
                        "appear \"off\" - to fix, set a config value for "
                        "[rattail] timezone.default")

    def _write_logging_config_file(self):

        # load all current values into configparser
        parser = configparser.RawConfigParser()
        for section, values in self.configuration.items():
            parser.add_section(section)
            for option, value in values.items():
                parser.set(section, option, value)

        # write INI file and return path
        path = temp_path(suffix='.conf')
        with open(path, 'wt') as f:
            parser.write(f)
        return path

    # TODO: use this or lose it etc.
    def beaker_invalidate_setting(self, name):
        # if not self.beaker_caching:
        #     return

        # # tell beaker to remove the cached value for this setting
        # self.beaker_config_cache.remove_value(key=name)

        pass


class RattailConfig(RattailConfigMixin):
    """
    .. note::

       This class has served us well over the years, but will soon be
       replaced with :class:`RattailConfiguration`.  See docs for that
       new class if you wish to enable it in your app.

    Rattail config object; this represents the sum total of configuration
    available to the running app.  The actual config available falls roughly
    into two categories: the "defaults" and the "db" (more on these below).
    The general idea here is that one might wish to provide some default
    settings within some config file(s) and/or the command line itself, but
    then allow all settings found in the database to override those defaults.
    However, all variations on this theme are supported, e.g. "use db settings
    but prefer those from file", "never use db settings", and so on.

    As for the "defaults" aspect of the config, this is read only once upon
    application startup.  It almost certainly involves one (or more) config
    file(s), but in addition to that, the application itself is free to embed
    default settings within the config object.  When this occurs, there will be
    no distinction made between settings which came from a file versus those
    which were established as defaults by the application logic.

    As for the "db" aspect of the config, of course this ultimately hinges upon
    the config defaults.  If a default Rattail database connection is defined,
    then the ``Setting`` table within that database may also be consulted for
    config values.  When this is done, the ``Setting.name`` is determined by
    concatenating the ``section`` and ``option`` arguments from the
    :meth:`get()` call, with a period (``'.'``) in between.
    """

    def __init__(self, files=[], usedb=None, preferdb=None):
        self.files_requested = []
        self.files_read = []
        self.parser = configparser.ConfigParser()
        for path in files:
            self.read_file(path)
        self.usedb = usedb
        if self.usedb is None:
            self.usedb = self.getbool('rattail.config', 'usedb', usedb=False, default=False)
        self.preferdb = preferdb
        if self.usedb and self.preferdb is None:
            self.preferdb = self.getbool('rattail.config', 'preferdb', usedb=False, default=False)

        # Attempt to detect lack of SQLAlchemy libraries etc.  This allows us
        # to avoid installing those on a machine which will not need to access
        # a database etc.
        if self.usedb:
            try:
                from rattail.db import Session
            except ImportError: # pragma: no cover
                log.warning("config created with `usedb = True`, but can't import "
                            "`rattail.db.Session`, so setting `usedb = False` instead",
                            exc_info=True)
                self.usedeb = False
                self.preferdb = False

        # this is always false, unless/until it becomes true
        self.versioning_has_been_enabled = False

        # should we use beaker caching?
        self.beaker_caching = self.getbool('rattail.config', 'beaker_cache.enabled',
                                           usedb=False, default=False)
        if self.beaker_caching:
            from beaker.cache import CacheManager
            from beaker.util import parse_cache_config_options

            # default cache settings
            namespace = None
            options = {
                'cache.type': 'file',
                'cache.data_dir': os.path.join(self.appdir(), 'cache',
                                               'config', 'data'),
                'cache.lock_dir': os.path.join(self.appdir(), 'cache',
                                               'config', 'lock'),
            }

            # but let config override and supplement
            if self.parser.has_section('rattail.config'):
                for option in self.parser.options('rattail.config'):
                    if option == 'beaker_cache.namespace':
                        namespace = self.parser.get('rattail.config', option)
                        continue
                    prefix = 'beaker_cache.'
                    if option.startswith(prefix):
                        key = 'cache.{}'.format(option[len(prefix):])
                        options[key] = self.parser.get('rattail.config', option)

            # and with those options, make a cache manager
            self.beaker_cache_manager = CacheManager(
                **parse_cache_config_options(options))

            # use app-specific namespace in case multiple apps use the
            # same (e.g. memcached) backend.  note that we're making a
            # "best faith" effort here; using the name of the python
            # environment root dir.
            if not namespace:
                appdir = self.appdir()
                rootdir = os.path.dirname(appdir)
                envname = os.path.basename(rootdir)
                namespace = 'rattail.config:{}'.format(envname)

            # and finally, make our cache
            self.beaker_config_cache = self.beaker_cache_manager.get_cache(
                namespace)

    @property
    def prioritized_files(self):
        return reversed(self.files_read)

    def read_file(self, path, recurse=True, require=False):
        """
        Read in config from the given file.

        By default this will "crawl" the file and recursively read in
        any other config files which this one "includes" (or
        "requires").

        :param path: Path to the config file to read.

        :param recurse: Whether the file should be recursively
           "crawled" as described above.

        :param require: If true, and the given file path is not
           readable, will raise an error.
        """
        path = os.path.abspath(path)
        if path in self.files_requested:
            log.debug("ignoring config file which was already requested: {0}".format(path))
            return

        log.debug("will attempt to read config from file: {0}".format(path))
        self.files_requested.append(path)

        parser = configparser.ConfigParser(dict(
            here=os.path.dirname(path),
        ))
        if not parser.read(path):
            log.debug("ConfigParser.read() failed")
            if require:
                raise RuntimeError("Cannot read from config file path: {}".format(path))
            return

        # If recursing, walk the complete config file inheritance chain.
        if recurse:
            if parser.has_section('rattail.config'):

                # first bring in any "required" files
                if parser.has_option('rattail.config', 'require'):
                    includes = self.parse_list(
                        parser.get('rattail.config', 'require'))
                    for included in includes:
                        self.read_file(included, recurse=True, require=True)

                # next try to bring in any "included" files
                if parser.has_option('rattail.config', 'include'):
                    includes = self.parse_list(
                        parser.get('rattail.config', 'include'))
                    for included in includes:
                        self.read_file(included, recurse=True)

        # Okay, now we can finally read this file into our main parser.
        self.parser.read(path)
        self.files_read.append(path)
        log.info("config was read from file: {0}".format(path))

    def configure_logging(self):
        """
        This first checks current config to determine whether or not we're
        supposed to be configuring logging at all.  If not, nothing more is
        done.

        If we are to configure logging, then this will save the current config
        parser defaults to a temporary file, and use this file to configure
        Python's standard logging module.
        """
        if not self.getbool('rattail.config', 'configure_logging', usedb=False, default=False):
            return

        # Coerce all logged timestamps to the local timezone, if possible.
        timezone = self.get('rattail', 'timezone.default', usedb=False)
        if timezone:
            logging.Formatter.converter = TimeConverter(self)

        # Flush all current config to a single file, for input to fileConfig().
        path = temp_path(suffix='.conf')
        with open(path, 'wt') as f:
            self.parser.write(f)

        try:
            logging.config.fileConfig(path, disable_existing_loggers=False)
        except configparser.NoSectionError as error:
            log.warning("tried to configure logging, but got NoSectionError: {0}".format(error))
        else:
            log.debug("configured logging")
        finally:
            os.remove(path)

        # if we could not coerce timestamps, log a warning about that
        if not timezone:
            log.warning("timezone is not set, so logging timestamps may "
                        "appear \"off\" - to fix, set a config value for "
                        "[rattail] timezone.default")

    def setdefault(self, section, option, value):
        """
        Establishes a new default for the given setting, if none exists yet.
        The effective default value is returned in all cases.
        """
        exists = True
        if not self.parser.has_section(section):
            self.parser.add_section(section)
            exists = False
        elif not self.parser.has_option(section, option):
            exists = False
        if not exists:
            self.parser.set(section, option, value)
        return self.parser.get(section, option)

    def set(self, section, option, value):
        """
        Set a value within the config's parser data set, i.e. the "defaults".
        This should probably be used sparingly, though one expected use is
        within tests (for convenience).
        """
        if not self.parser.has_section(section):
            self.parser.add_section(section)
        self.parser.set(section, option, value)

    def get(self, section, option, usedb=None, preferdb=None, session=None, default=None):
        """
        Retrieve a value from config.
        """
        usedb = usedb if usedb is not None else self.usedb
        if usedb:
            preferdb = preferdb if preferdb is not None else getattr(self, 'preferdb', False)
        else:
            preferdb = False
        
        if usedb and preferdb:
            value = self._getdb(section, option, session=session)
            if value is not None:
                return value

        if self.parser.has_option(section, option):
            return self.parser.get(section, option)

        if usedb and not preferdb:
            value = self._getdb(section, option, session=session)
            if value is not None:
                return value

        return default

    def _getdb(self, section, option, session=None):
        """
        Retrieve a config value from database settings table.
        """
        app = self.get_app()
        name = '{}.{}'.format(section, option)

        if self.beaker_caching:
            return self._getdb_beaker(section, option)

        close = False
        if not session:
            # nb. avoid continuum user lookup
            session = app.make_session(continuum_user=None)
            close = True

        value = app.get_setting(session, name)

        if close:
            session.close()

        return value

    def _getdb_beaker(self, section, option):
        """
        Retrieve a config value from database settings table.
        """
        app = self.get_app()
        name = '{}.{}'.format(section, option)

        def get_value():
            # nb. avoid continuum user lookup
            session = app.make_session(continuum_user=None)
            value = app.get_setting(session, name)
            session.close()
            log.debug("had to (re-)fetch setting '%s' from db: %s", name, value)
            return value

        return self.beaker_config_cache.get(key=name, createfunc=get_value)

    def beaker_invalidate_setting(self, name):
        if not self.beaker_caching:
            return

        # tell beaker to remove the cached value for this setting
        self.beaker_config_cache.remove_value(key=name)

    def setdb(self, section, option, value, session=None):
        """
        Set a config value in the database settings table.  Note that the
        ``value`` arg should be a Unicode object.
        """
        app = self.get_app()

        close = False
        if not session:
            # nb. avoid continuum user lookup
            session = app.make_session(continuum_user=None)
            close = True

        name = '{}.{}'.format(section, option)
        app.save_setting(session, name, value)

        if close:
            session.commit()
            session.close()

    def getbool(self, *args, **kwargs):
        """
        Retrieve a boolean value from config.
        """
        value = self.get(*args, **kwargs)
        return parse_bool(value)

    def getint(self, *args, **kwargs):
        """
        Retrieve an integer value from config.
        """
        value = self.get(*args, **kwargs)
        if value is None:
            return None
        if value == '' and 'default' in kwargs:
            return kwargs['default']
        if isinstance(value, int):
            return value
        return int(value)

    def getdate(self, *args, **kwargs):
        """
        Retrieve a date value from config.
        """
        value = self.get(*args, **kwargs)
        app = self.get_app()
        return app.parse_date(value)

    def getlist(self, *args, **kwargs):
        """
        Retrieve a list of string values from a single config option.
        """
        value = self.get(*args, **kwargs)
        if value is None:
            return None
        if isinstance(value, str):
            return self.parse_list(value)
        return value            # maybe a caller-provided default?

    def parse_bool(self, value):
        return parse_bool(value)

    def parse_list(self, value):
        return parse_list(value)

    def make_list_string(self, values):
        """
        Coerce the given list of values to a string, for config
        storage.  If this string is later parsed via
        :meth:`parse_list()` then it should return the same list of
        values.
        """
        final = []
        for value in values:
            if ' ' in value:
                quote = '"' if "'" in value else "'"
                value = f"{quote}{value}{quote}"
            final.append(value)
        return ', '.join(final)

    def get_dict(self, section):
        """
        Convenience method which returns a dictionary of options contained
        within the given section.  Note that this method only supports the
        "default" config settings, i.e. those within the underlying parser.
        """
        settings = {}
        if self.parser.has_section(section):
            for option in self.parser.options(section):
                settings[option] = self.parser.get(section, option)
        return settings

    def require(self, section, option, **kwargs):
        """
        Fetch a value from current config, and raise an error if no value can
        be found.
        """
        if 'default' in kwargs:
            warnings.warn("You have provided a default value to the `RattailConfig.require()` "
                          "method.  This is allowed but also somewhat pointless, since `get()` "
                          "would suffice if a default is known.", UserWarning)

        msg = kwargs.pop('msg', None)
        value = self.get(section, option, **kwargs)
        if value is not None:
            return value

        if msg is None:
            msg = "Missing or invalid config"
        msg = "{0}; please set '{1}' in the [{2}] section of your config file".format(
            msg, option, section)
        raise ConfigurationError(msg)

    ##############################
    # deprecated methods
    ##############################

    def options(self, section):
        warnings.warn("RattailConfig.option() is deprecated, please find "
                      "another way to accomplish what you're after.",
                      DeprecationWarning, stacklevel=2)
        return self.parser.options(section)

    def has_option(self, section, option):
        warnings.warn("RattailConfig.has_option() is deprecated, please find "
                      "another way to accomplish what you're after.",
                      DeprecationWarning, stacklevel=2)
        return self.parser.has_option(section, option)


class ConfigExtension(object):
    """
    Base class for all config extensions.
    """
    key = None

    def __repr__(self):
        return "ConfigExtension(key={0})".format(repr(self.key))

    def configure(self, config):
        """
        All subclasses should override this method, to extend the config object
        in any way necessary etc.
        """


def make_config(
        files=None,
        plus_files=None,
        usedb=None,
        preferdb=None,
        env=None,
        winsvc=None,
        extend=True,
        versioning=None):
    """
    Returns a new config object, initialized with the given parameters and
    further modified by all registered config extensions.

    :param files: Config path(s) to be loaded.  If not specified, then some
       "default" behavior will be attempted.  This will check for env var
       ``RATTAIL_CONFIG_FILES`` or fallback to system default paths.  Or
       you can override all that by specifying some path(s) here.

    :param plus_files: Additional config path(s) to be loaded.  You
       may specify a "config tweak" file here, and leave ``files``
       empty, to get "normal plus tweak" behavior.

    :param versioning: Controls whether or not the versioning system is
       configured with the new config object.  If ``True``, versioning will be
       configured.  If ``False`` then it will not be configured.  If ``None``
       (the default) then versioning will be configured only if the config
       object itself says that it should be.
    """
    if env is None:
        env = os.environ
    if files is None:
        files = env.get('RATTAIL_CONFIG_FILES')
        if files is not None:
            files = files.split(os.pathsep)
        else:
            # try to guess a default config path
            # TODO: for now, prefer app/quiet.conf if present, but
            # probably we should look for adhoc.conf instead, since
            # the point of this magic is to make running ad-hoc
            # commands easier..
            quiet = os.path.join(sys.prefix, 'app', 'quiet.conf')
            if os.path.exists(quiet):
                # this config is definitely app-specific
                files = [quiet]
            else:
                # generic fallback, probably not very useful
                files = default_system_paths() + default_user_paths()
    elif isinstance(files, str):
        files = [files]

    if plus_files:
        if isinstance(plus_files, str):
            plus_files = [plus_files]
        files += plus_files

    # If making config for a Windows service, we must read the default config
    # file(s) first, and check it to see if there is an alternate config file
    # which should be considered the "root" file.  Normally we specify the root
    # config file(s) via command line etc., but there is no practical way to
    # pass parameters to a Windows service.  This way we can effectively do
    # just that, via config.
    if winsvc is not None:
        parser = configparser.SafeConfigParser()
        parser.read(files)
        if parser.has_section('rattail.config'):
            key = 'winsvc.{0}'.format(winsvc)
            if parser.has_option('rattail.config', key):
                files = parse_list(parser.get('rattail.config', key))

    # Initial config object will have values read from the given file paths,
    # and kwargs, but no other app defaults etc. will have been applied yet.
    config = RattailConfigWrapper(files, usedb=usedb, preferdb=preferdb)
    log.debug("using config object of type: %s",
              config.config.__class__.__name__)
    log.debug("config files were: %s", config.files_read)

    # turn on display of rattail deprecation warnings by default
    # TODO: this should be configurable, and possibly live elsewhere?
    warnings.filterwarnings('default', category=DeprecationWarning,
                            module=r'^rattail')
    warnings.filterwarnings('default', category=DeprecationWarning,
                            module=r'^tailbone')

    if config.getbool('rattail', 'suppress_psycopg2_wheel_warning', usedb=False):
        # TODO: revisit this, does it require action from us?
        # suppress this warning about psycopg2 wheel; not sure what it means yet
        # exactly but it's causing frequent noise for us...
        warnings.filterwarnings(
            'ignore',
            r'^The psycopg2 wheel package will be renamed from release 2\.8; in order to keep '
            r'installing from binary please use "pip install psycopg2-binary" instead\. For details '
            r'see: <http://initd.org/psycopg/docs/install.html#binary-install-from-pypi>\.',
            UserWarning,
            r'^psycopg2$',
        )

    # Apply extra config for all available extensions.
    if extend:
        extensions = load_entry_points('rattail.config.extensions')
        for extension in extensions.values():
            log.debug("applying '{0}' config extension".format(extension.key))
            extension().configure(config)

    # maybe configure versioning
    if versioning is None:
        versioning = config.versioning_enabled()
    if versioning:
        from rattail.db.config import configure_versioning
        configure_versioning(config)

    # maybe set "future" behavior for SQLAlchemy
    if config.getbool('rattail.db', 'sqlalchemy_future_mode', usedb=False):
        from rattail.db import Session
        if Session:
            Session.configure(future=True)

    return config


def default_system_paths():
    """
    Returns a list of default system-level config file paths, according to the
    current platform.
    """
    if sys.platform == 'win32':

        # Use the Windows Extensions libraries to fetch official defaults.
        try:
            from win32com.shell import shell, shellcon
        except ImportError:
            # raise WindowsExtensionsNotInstalled
            return []
        else:
            return [
                os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_COMMON_APPDATA), 'rattail.conf'),
                os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_COMMON_APPDATA), 'rattail', 'rattail.conf'),
            ]

    return [
        '/etc/rattail.conf',
        '/etc/rattail/rattail.conf',
        '/usr/local/etc/rattail.conf',
        '/usr/local/etc/rattail/rattail.conf',
    ]


def default_user_paths():
    """
    Returns a list of default user-level config file paths, according to the
    current platform.
    """
    if sys.platform == 'win32':

        # Use the Windows Extensions libraries to fetch official defaults.
        try:
            from win32com.shell import shell, shellcon
        except ImportError:
            # raise WindowsExtensionsNotInstalled
            return []
        else:
            return [
                os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_APPDATA), 'rattail.conf'),
                os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_APPDATA), 'rattail', 'rattail.conf'),
            ]

    return [
        os.path.expanduser('~/.rattail.conf'),
        os.path.expanduser('~/.rattail/rattail.conf'),
    ]


def get_user_dir(create=False):
    """
    Returns a path to the "preferred" user-level folder, in which additional
    config files (etc.) may be placed as needed.  This essentially returns a
    platform-specific variation of ``~/.rattail/``.

    If ``create`` is ``True``, then the folder will be created if it does not
    already exist.
    """
    if sys.platform == 'win32':

        # Use the Windows Extensions libraries to fetch official defaults.
        try:
            from win32com.shell import shell, shellcon
        except ImportError:
            raise WindowsExtensionsNotInstalled
        else:
            path = os.path.join(shell.SHGetSpecialFolderPath(
                0, shellcon.CSIDL_APPDATA), 'rattail')

    else:
        path = os.path.expanduser('~/.rattail')

    if create and not os.path.exists(path):
        os.mkdir(path)
    return path


def get_user_file(filename, createdir=False):
    """
    Returns a full path to a user-level config file location.  This is obtained
    by first calling :func:`get_user_dir()` and then joining the result with
    ``filename``.

    The ``createdir`` argument will be passed to :func:`get_user_dir()` as its
    ``create`` arg, and may be used to ensure the user-level folder exists.
    """
    return os.path.join(get_user_dir(create=createdir), filename)


class ConfigProfile(object):
    """
    Generic class to represent a config "profile", as used by the filemon and
    datasync daemons, etc.

    .. todo::

       This clearly needs more documentation.

    .. attribute:: config

       Reference to the primary Rattail config object for the running app.

    .. attribute:: key

       String identifier unique to this profile, within the broader
       config section.
    """

    def __init__(self, config, key, **kwargs):
        self.config = config
        self.app = self.config.get_app()
        self.model = self.config.get_model()
        self.enum = self.config.get_enum()
        self.key = key
        self.prefix = kwargs.pop('prefix', key)
        self.load()

    def load(self):
        """
        Read all relevant settings etc. from the config object,
        setting attributes on this profile instance as needed.
        """

    def load_defaults(self):
        """
        Read all "default" (common) settings from config, for the
        current profile.
        """
        self.workdir = self._config_string('workdir')
        self.stop_on_error = self._config_boolean('stop_on_error', False)

    def load_actions(self):
        """
        Read the "actions" from config, for the current profile, and
        assign the result to ``self.actions``.
        """
        self.actions = []
        for action in self._config_list('actions'):
            self.actions.append(self._config_action(action))

    @property
    def section(self):
        """
        Each subclass of ``ConfigProfile`` must define this.
        """
        raise NotImplementedError

    def _config_string(self, option, **kwargs):
        return self.config.get(self.section,
                               '{}.{}'.format(self.prefix, option),
                               **kwargs)

    def _config_boolean(self, option, default=None):
        return self.config.getbool(self.section,
                                   '{}.{}'.format(self.prefix, option),
                                   default=default)

    def _config_int(self, option, minimum=1, default=None):
        """
        Retrieve the *integer* value for the given option.
        """
        option = '{}.{}'.format(self.prefix, option)

        # try to read value from config
        value = self.config.getint(self.section, option)
        if value is not None:

            # found a value; validate it
            if value < minimum:
                log.warning("config value %s is too small; falling back to minimum "
                            "of %s for option: %s", value, minimum, option)
                value = minimum

        # or, use default value, if valid
        elif default is not None and default >= minimum:
            value = default

        # or, just use minimum value
        else:
            value = minimum

        return value

    def _config_list(self, option, default=None, **kwargs):
        value = self._config_string(option, **kwargs)
        if value:
            return self.config.parse_list(value)

        if isinstance(default, list):
            return default

        return []

    def _config_action(self, name):
        """
        Retrieve an "action" value from config, for the current
        profile.  This returns a :class:`ConfigProfileAction`
        instance.
        """
        from rattail.monitoring import CommandAction

        function = self._config_string('action.{}.func'.format(name))
        class_ = self._config_string('action.{}.class'.format(name))
        cmd = self._config_string('action.{}.cmd'.format(name))

        specs = [1 if spec else 0 for spec in (function, class_, cmd)]
        if sum(specs) != 1:
            raise ConfigurationError(
                "Monitor profile '{}' (action '{}') must have exactly one of: "
                "function, class, command".format(self.prefix, name))

        action = ConfigProfileAction()
        action.config = self.config

        if function:
            action.spec = function
            action.action = load_object(action.spec)
        elif class_:
            action.spec = class_
            action.action = load_object(action.spec)(self.config)
        elif cmd:
            action.spec = cmd
            action.action = CommandAction(self.config, cmd)

        action.args = self._config_list('action.{}.args'.format(name))

        action.kwargs = {}
        pattern = re.compile(r'^{}\.action\.{}\.kwarg\.(?P<keyword>\w+)$'.format(self.prefix, name), re.IGNORECASE)
        settings = self.config.get_dict(self.section)
        for key in settings:
            match = pattern.match(key)
            if match:
                action.kwargs[match.group('keyword')] = settings[key]

        action.retry_attempts = self._config_int('action.{}.retry_attempts'.format(name), minimum=1)
        action.retry_delay = self._config_int('action.{}.retry_delay'.format(name), minimum=0)
        return action


class ConfigProfileAction(object):
    """
    Simple class to hold configuration for a particular "action"
    defined within a monitor :class:`ConfigProfile`.  Each instance
    has the following attributes:

    .. attribute:: spec

       The original "spec" string used to obtain the action callable.

    .. attribute:: action

       A reference to the action callable.

    .. attribute:: args

       A sequence of positional arguments to be passed to the callable
       (in addition to the file path) when invoking the action.

    .. attribute:: kwargs

       A dictionary of keyword arguments to be passed to the callable
       (in addition to the positional arguments) when invoking the
       action.

    .. attribute:: retry_attempts

       Number of attempts to make when invoking the action.  Defaults
       to ``1``, meaning the first attempt will be made but no retries
       will happen.

    .. attribute:: retry_delay

       Number of seconds to pause between retry attempts, if
       :attr:`retry_attempts` is greater than one.  Defaults to ``0``.
    """
    spec = None
    action = None
    args = []
    kwargs = {}
    retry_attempts = 1
    retry_delay = 0


class FreeTDSLoggingFilter(logging.Filter):
    """
    Custom logging filter, to suppress certain "write to server failed"
    messages relating to FreeTDS database connections.  They seem harmless and
    just cause unwanted error emails.
    """

    def __init__(self, *args, **kwargs):
        logging.Filter.__init__(self, *args, **kwargs)
        self.pattern = re.compile(r'(?:Read from|Write to) the server failed')

    def filter(self, record):
        if (record.name == 'sqlalchemy.pool.QueuePool'
            and record.funcName == '_finalize_fairy'
            and record.levelno == logging.ERROR
            and record.msg == "Exception during reset or similar"
            and record.exc_info
            and self.pattern.search(str(record.exc_info[1]))):

            # Log this as a warning instead of error, to cut down on our noise.
            record.levelno = logging.WARNING
            record.levelname = 'WARNING'

        return True



# -*- coding: utf-8 -*-
"""grow change.class_name

Revision ID: 6e6d6e2ef0e
Revises: 27bcaa3c7f5f
Create Date: 2016-02-01 11:22:08.833707

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '6e6d6e2ef0e'
down_revision = u'27bcaa3c7f5f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # change
    op.alter_column('change', 'class_name', type_=sa.String(length=40))


def downgrade():

    # change
    op.alter_column('change', 'class_name', type_=sa.String(length=25))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor invoice parser for KeHE Distributors
"""

import re
import csv
import datetime
import decimal

from rattail.vendors.invoices import InvoiceParser


class KeheInvoiceParser(InvoiceParser):
    """
    Vendor invoice parser for KeHE Distributors.

    KeHE has changed their format a few times; hence this parser is
    capable of handling a few similar but different formats.  All are
    essentially text/csv.

    Format #1 is the original and is tab-separated.

    Format #2 came circa 2020-02-11 and is comma-separated.

    Format #3 is like #1 but changed a column header.

    Format #4 is like #2 but changed column headers and data formats.

    See the :meth:`detect_version()` method for details of how the
    parser decides which "version" (format) of file it's dealing with.
    """
    key = 'rattail.contrib.kehe'
    display = "KeHE Distributors"
    vendor_key = 'kehe'

    pack_size_pattern = re.compile('^(?P<case_quantity>\d+)/(?P<size>\d*\.\d+ \w\w)$')
    bulk_size_pattern = re.compile('^(?P<case_quantity>\d+\.\d+) LB$')

    def detect_version(self, path):
        """
        This will inspect the data file and return the "version" of format it
        thinks we're dealing with.
        """
        with open(path, 'rt') as f:
            line = f.readline()

        delimiter = str('\t') if '\t' in line else ','

        csv_file = open(path, 'rt')
        reader = csv.DictReader(csv_file, delimiter=delimiter)
        data = next(reader)
        csv_file.close()

        if delimiter == ',': # comma-separated
            if 'OrderQuatity' in data:
                return 4
            return 2

        else: # tab-separated
            if 'InvoiceDate' in data:
                return 3
            return 1

    def parse_invoice_number(self, path):

        # first find out which version of file we have
        version = self.detect_version(path)

        delimiter = str('\t') if version in (1, 3) else ','
        csv_file = open(path, 'rt')
        reader = csv.DictReader(csv_file, delimiter=delimiter)
        data = next(reader)
        csv_file.close()

        return data['InvoiceNumber']

    def parse_invoice_date(self, path):

        # first find out which version of file we have
        version = self.detect_version(path)

        delimiter = str('\t') if version in (1, 3) else ','
        csv_file = open(path, 'rt')
        reader = csv.DictReader(csv_file, delimiter=delimiter)
        data = next(reader)
        csv_file.close()

        if version == 1:
            return datetime.datetime.strptime(data['Invoice Date'], '%Y-%m-%d').date()
        elif version in (2, 3):
            try:
                return datetime.datetime.strptime(data['InvoiceDate'], '%m/%d/%Y %I:%M:%S %p').date()
            except ValueError:
                return datetime.datetime.strptime(data['InvoiceDate'], '%m/%d/%Y').date()
        elif version == 4:
            return datetime.datetime.strptime(data['InvoiceDate'], '%m/%d/%Y').date()

    def parse_rows(self, path):

        # first find out which version of file we have
        version = self.detect_version(path)
        delimiter = str('\t') if version in (1, 3) else ','

        # default fields for old version 1
        fields = {
            'upc': 'UPC Code',
            'ship_item': 'Ship Item',
            'brand': 'Brand',
            'description': 'Description',
            'order_quantity': 'Order Qty',
            'ship_quantity': 'Ship Qty',
            'net_each': 'Net Each',
            'net_billable': 'Net Billable',
            'pack_size': 'Pack Size',
        }

        if version > 1:
            fields.update({
                'upc': 'Upc',
                'ship_item': 'ShipItem',
                'order_quantity': 'OrderQuantity',
                'ship_quantity': 'ShipQuantity',
                'net_each': 'NetEach',
                'net_billable': 'NetBillable',
                'pack_size': 'PackSize',
            })

        if version == 4:
            fields.update({
                'order_quantity': 'OrderQuatity',
            })

        csv_file = open(path, 'rt')
        reader = csv.DictReader(csv_file, delimiter=delimiter)

        for i, data in enumerate(reader, 1):

            row = self.make_row()
            row.line_number = i
            row.item_entry = data[fields['upc']]
            row.upc = self.app.make_gpc(row.item_entry)
            row.vendor_code = data[fields['ship_item']]
            row.brand_name = data[fields['brand']]
            row.description = data[fields['description']]
            row.ordered_units = self.int_(data[fields['order_quantity']])
            row.shipped_units = self.int_(data[fields['ship_quantity']])
            row.total_cost = self.decimal(data[fields['net_billable']])

            # nb. unit cost ideally would come from NetEach column,
            # but that is rounded to 2 places so not always accurate
            if row.total_cost and row.shipped_units:
                row.unit_cost = row.total_cost / row.shipped_units
            else:
                row.unit_cost = self.decimal(data[fields['net_each']])

            # Case quantity may be embedded in size string.
            row.size = data[fields['pack_size']]
            row.case_quantity = 1
            match = self.pack_size_pattern.match(row.size)
            if match:
                row.case_quantity = int(match.group('case_quantity'))
                row.size = match.group('size')

                if row.case_quantity == 1:
                    match = self.bulk_size_pattern.match(row.size)
                    if match:
                        row.case_quantity = decimal.Decimal(match.group('case_quantity'))
                        row.unit_cost /= row.case_quantity
                        row.ordered_cases = row.ordered_units
                        row.ordered_units = None
                        row.shipped_cases = row.shipped_units
                        row.shipped_units = None

            # attach original raw data to the row we're returning; caller
            # can use as needed, or ignore
            row._raw_data = data

            yield row

        csv_file.close()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
`Product Open Data`_ Integration

.. _`Product Open Data`: http://www.product-open-data.com/
"""

import os


def get_image_path(config, gpc):
    """
    Get an image file path from a product GPC.
    """
    root_path = config.get('rattail.pod', 'pictures.gtin.root_path')
    if root_path:
        gtin = str(gpc)[1:]
        return os.path.join(root_path, 'gtin-{}'.format(gtin[:3]), '{}.jpg'.format(gtin))


def get_image_url(config, gpc, require=True):
    """
    Get an image URL from a product GPC.
    """
    if gpc:
        getter = config.require if require else config.get
        root_url = getter('rattail.pod', 'pictures.gtin.root_url')
        if root_url:
            root_url = root_url.rstrip('/')
            gtin = str(gpc)[1:]
            return '{}/gtin-{}/{}.jpg'.format(root_url, gtin[:3], gtin)

    return config.get('rattail.pod', 'pictures.gtin.not_found_url')



# -*- coding: utf-8 -*-
"""add members

Revision ID: f3781f071de3
Revises: 45b90e4f8ab4
Create Date: 2018-12-19 21:46:02.335362

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'f3781f071de3'
down_revision = '45b90e4f8ab4'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # member
    op.create_table('member',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.String(length=20), nullable=True),
                    sa.Column('person_uuid', sa.String(length=32), nullable=True),
                    sa.Column('customer_uuid', sa.String(length=32), nullable=True),
                    sa.Column('joined', sa.Date(), nullable=True),
                    sa.Column('active', sa.Boolean(), nullable=False),
                    sa.Column('equity_current', sa.Boolean(), nullable=True),
                    sa.Column('equity_total', sa.Numeric(precision=5, scale=2), nullable=True),
                    sa.Column('equity_payment_due', sa.Date(), nullable=True),
                    sa.Column('equity_last_paid', sa.Date(), nullable=True),
                    sa.Column('equity_payment_credit', sa.Numeric(precision=5, scale=2), nullable=False),
                    sa.Column('withdrew', sa.Date(), nullable=True),
                    sa.ForeignKeyConstraint(['customer_uuid'], ['customer.uuid'], name='member_fk_customer'),
                    sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'], name='member_fk_person'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('member_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('person_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('customer_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('joined', sa.Date(), autoincrement=False, nullable=True),
                    sa.Column('active', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('equity_current', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('equity_total', sa.Numeric(precision=5, scale=2), autoincrement=False, nullable=True),
                    sa.Column('equity_payment_due', sa.Date(), autoincrement=False, nullable=True),
                    sa.Column('equity_last_paid', sa.Date(), autoincrement=False, nullable=True),
                    sa.Column('equity_payment_credit', sa.Numeric(precision=5, scale=2), autoincrement=False, nullable=True),
                    sa.Column('withdrew', sa.Date(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_member_version_end_transaction_id'), 'member_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_member_version_operation_type'), 'member_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_member_version_transaction_id'), 'member_version', ['transaction_id'], unique=False)


def downgrade():

    # member
    op.drop_index(op.f('ix_member_version_transaction_id'), table_name='member_version')
    op.drop_index(op.f('ix_member_version_operation_type'), table_name='member_version')
    op.drop_index(op.f('ix_member_version_end_transaction_id'), table_name='member_version')
    op.drop_table('member_version')
    op.drop_table('member')



# -*- coding: utf-8 -*-
"""add productcost.discontinued

Revision ID: 6a1ec8b93637
Revises: 3b5e0b87c176
Create Date: 2017-03-28 18:02:44.714003

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '6a1ec8b93637'
down_revision = u'3b5e0b87c176'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # product_cost
    op.add_column('product_cost', sa.Column('discontinued', sa.Boolean(), nullable=True))


def downgrade():

    # product_cost
    op.drop_column('product_cost', 'discontinued')



# -*- coding: utf-8; -*-
"""add pending_product

Revision ID: 43b9e0a6c14e
Revises: 8856f697902d
Create Date: 2021-11-09 18:08:48.252950

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '43b9e0a6c14e'
down_revision = '8856f697902d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # pending_product
    op.create_table('pending_product',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('user_uuid', sa.String(length=32), nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('upc', rattail.db.types.GPCType(), nullable=True),
                    sa.Column('scancode', sa.String(length=14), nullable=True),
                    sa.Column('item_id', sa.String(length=50), nullable=True),
                    sa.Column('item_type', sa.Integer(), nullable=True),
                    sa.Column('department_name', sa.String(length=30), nullable=True),
                    sa.Column('department_uuid', sa.String(length=32), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('brand_uuid', sa.String(length=32), nullable=True),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('size', sa.String(length=30), nullable=True),
                    sa.Column('case_size', sa.Numeric(precision=9, scale=4), nullable=True),
                    sa.Column('regular_price_amount', sa.Numeric(precision=8, scale=3), nullable=True),
                    sa.Column('special_order', sa.Boolean(), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(['brand_uuid'], ['brand.uuid'], name='pending_product_fk_brand'),
                    sa.ForeignKeyConstraint(['department_uuid'], ['department.uuid'], name='pending_product_fk_department'),
                    sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'], name='pending_product_fk_user'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # pending_product
    op.drop_table('pending_product')



# -*- coding: utf-8; -*-
"""add pricing_batch.min_diff_percent

Revision ID: 7eb0903932ff
Revises: d28d28bc3cd0
Create Date: 2018-11-25 19:47:21.873271

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '7eb0903932ff'
down_revision = 'd28d28bc3cd0'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_pricing
    op.add_column('batch_pricing', sa.Column('min_diff_percent', sa.Numeric(precision=5, scale=2), nullable=True))


def downgrade():

    # batch_pricing
    op.drop_column('batch_pricing', 'min_diff_percent')



# -*- coding: utf-8; -*-
"""add PendingProduct.product

Revision ID: 86b00e66b83e
Revises: a7c2d5d84743
Create Date: 2023-09-09 21:37:39.905677

"""

# revision identifiers, used by Alembic.
revision = '86b00e66b83e'
down_revision = 'a7c2d5d84743'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # pending_product
    op.add_column('pending_product', sa.Column('product_uuid', sa.String(length=32), nullable=True))
    op.add_column('pending_product', sa.Column('resolved', sa.DateTime(), nullable=True))
    op.add_column('pending_product', sa.Column('resolved_by_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('pending_product_fk_product', 'pending_product', 'product', ['product_uuid'], ['uuid'])
    op.create_foreign_key('pending_product_fk_resolved_by', 'pending_product', 'user', ['resolved_by_uuid'], ['uuid'])


def downgrade():

    # pending_product
    op.drop_constraint('pending_product_fk_resolved_by', 'pending_product', type_='foreignkey')
    op.drop_constraint('pending_product_fk_product', 'pending_product', type_='foreignkey')
    op.drop_column('pending_product', 'resolved_by_uuid')
    op.drop_column('pending_product', 'resolved')
    op.drop_column('pending_product', 'product_uuid')



# -*- coding: utf-8; -*-
"""normalize percentages

Revision ID: ffd65c52b0fd
Revises: 7d009a925f21
Create Date: 2022-10-29 13:06:49.034231

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'ffd65c52b0fd'
down_revision = '7d009a925f21'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types


product_volatile = sa.sql.table('product_volatile',
                                sa.sql.column('true_margin'))


def convert_all(convert):

    # product_volatile
    op.execute(product_volatile.update()\
               .where(product_volatile.c.true_margin != None)\
               .values({'true_margin': convert(product_volatile.c.true_margin)}))


def upgrade():

    # product_volatile
    op.alter_column('product_volatile', 'true_margin', type_=sa.Numeric(precision=9, scale=5))

    # batch_pricing_row
    op.alter_column('batch_pricing_row', 'old_true_margin', type_=sa.Numeric(precision=9, scale=5))
    op.alter_column('batch_pricing_row', 'true_margin', type_=sa.Numeric(precision=9, scale=5))

    # input values are 0.0 - 1.0 range, output is 0 - 100
    convert_all(lambda val: val * 100)


def downgrade():

    # input values are 0 - 100 range, output is 0.0 - 1.0
    convert_all(lambda val: val / 100.0)

    # batch_pricing_row
    op.alter_column('batch_pricing_row', 'old_true_margin', type_=sa.Numeric(precision=8, scale=3))
    op.alter_column('batch_pricing_row', 'true_margin', type_=sa.Numeric(precision=8, scale=3))

    # product_volatile
    op.alter_column('product_volatile', 'true_margin', type_=sa.Numeric(precision=8, scale=5))
